#!/usr/bin/env python
# pylint: disable=R0903
"""
_StoreFail_

Util class to stage out a set of files from a job report to the /store/unmerged/fail
namespace

"""
from __future__ import print_function
from builtins import object

from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutMgr import StageOutMgr


def modifyLFN(inputLfn):
    """
    _modifyLFN_

    Util to tweak a normal LFN to be a /store/unmerged/fail LFN
    Simple algorithm to start with, split of /store/whatever and replace
    it with /store/unmerged/fail

    """
    lfnSplit = [x for x in inputLfn.split("/") if len(x) != 0]
    lfnSplit[1] = "unmerged/fail"
    newLfn = "/%s" % "/".join(lfnSplit)
    return newLfn


class StoreFailMgr(object):
    """
    _StoreFailMgr_

    """

    def __init__(self, jobReport):
        self.report = jobReport
        self.mgr = StageOutMgr()
        self.failures = {}

    def __call__(self):
        """
        _operator()_

        Invoke stage out of files to /store/unmerged/fail based on
        information in the report provided.

        Generate a list of the LFNs that were staged out as a
        return value

        """
        stagedOutFiles = []
        for fileToStage in self.report.files:
            lfn = fileToStage['LFN']
            newLfn = modifyLFN(lfn)
            print("Remapping LFN:\n%s\n" % lfn)
            print(" -TO-\n%s\n" % newLfn)
            print("Staging Out: %s" % newLfn)
            fileToStage['LFN'] = newLfn
            try:
                result = self.mgr(**fileToStage)
                fileToStage.update(result)
                stagedOutFiles.append(newLfn)
            except StageOutFailure as ex:
                msg = "Unable to stage out %s " % newLfn
                msg += " due to Stage Out Failure:\n"
                msg += str(ex)
                self.failures[newLfn] = msg
                continue

        return stagedOutFiles



#!/usr/bin/env python

"""
JobTracker test
"""
from __future__ import print_function, division
from builtins import range


import os
import os.path
import threading
import unittest
import stat
import subprocess
import getpass
import time
import cProfile
import pstats
from nose.plugins.attrib import attr

import WMCore.WMInit
from WMCore.DAOFactory    import DAOFactory
from WMCore.Services.UUIDLib import makeUUID

from WMCore.WMBS.File         import File
from WMCore.WMBS.Fileset      import Fileset
from WMCore.WMBS.Workflow     import Workflow
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.JobGroup     import JobGroup
from WMCore.WMBS.Job          import Job

from WMCore.DataStructs.Run   import Run

from WMComponent.JobTracker.JobTrackerPoller import JobTrackerPoller
from WMCore.ResourceControl.ResourceControl  import ResourceControl
from WMCore.JobStateMachine.ChangeState      import ChangeState
from WMQuality.TestInitCouchApp import TestInitCouchApp as TestInit
from WMQuality.Emulators import EmulatorSetup
from WMQuality.Emulators.EmulatedUnitTestCase import EmulatedUnitTestCase


def createJDL(jooID, directory, jobCE):
    """
    _createJDL_

    Create a simple JDL string list
    """

    jdl = []

    jdl.append("universe = globus\n")
    jdl.append("should_transfer_executable = TRUE\n")
    jdl.append("notification = NEVER\n")
    jdl.append("Executable = %s/submit.sh\n" % (directory))
    jdl.append("Output = condor.$(Cluster).$(Process).out\n")
    jdl.append("Error = condor.$(Cluster).$(Process).err\n")
    jdl.append("Log = condor.$(Cluster).$(Process).log\n")
    jdl.append("initialdir = %s\n" % directory)
    jdl.append("globusscheduler = %s\n" % (jobCE))
    jdl.append("+WMAgent_JobID = %s\n" % jooID)
    jdl.append("+WMAgent_AgentName = testAgent\n")
    jdl.append("Queue 1")
    return jdl


def createSubmitScript(directory):
    """
    _createSubmitScript_

    Create a stupid submit script
    """


    script = """
    #!/bin/bash

    sleep 900

    exit 0
    """

    path = os.path.join(directory, 'submit.sh')

    f = open(path, 'w')
    f.write(script)
    f.close()

    os.chmod(path,
             stat.S_IXGRP | stat.S_IXOTH | stat.S_IXUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR | stat.S_IWGRP)

    return


def getCondorRunningJobs(user):
    """
    _getCondorRunningJobs_

    Return the number of jobs currently running for a user
    """


    command = ['condor_q', user]
    pipe = subprocess.Popen(command, stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE, shell=False)
    stdout, error = pipe.communicate()

    output = stdout.split('\n')[-2]

    nJobs = int(output.split(';')[0].split()[0])

    return nJobs


class JobTrackerTest(EmulatedUnitTestCase):
    """
    TestCase for TestJobTracker module
    """

    _maxMessage = 10

    def setUp(self):
        """
        setup for test.
        """
        super(JobTrackerTest, self).setUp()
        myThread = threading.currentThread()

        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        # self.testInit.clearDatabase(modules = ["WMCore.WMBS", "WMCore.BossAir", "WMCore.ResourceControl"])
        self.testInit.setSchema(customModules=["WMCore.WMBS", "WMCore.BossAir", "WMCore.ResourceControl"],
                                useDefault=False)
        self.testInit.setupCouch("jobtracker_t/jobs", "JobDump")
        self.testInit.setupCouch("jobtracker_t/fwjrs", "FWJRDump")

        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)
        self.getJobs = self.daoFactory(classname="Jobs.GetAllJobs")


        # Create sites in resourceControl
        resourceControl = ResourceControl()
        resourceControl.insertSite(siteName='malpaquet', pnn='se.malpaquet',
                                   ceName='malpaquet', plugin="CondorPlugin")
        resourceControl.insertThreshold(siteName='malpaquet', taskType='Processing', \
                                        maxSlots=10000, pendingSlots=10000)

        locationAction = self.daoFactory(classname="Locations.New")
        locationAction.execute(siteName="malpaquet", pnn="malpaquet",
                               ceName="malpaquet", plugin="CondorPlugin")

        # Create user
        newuser = self.daoFactory(classname="Users.New")
        newuser.execute(dn="jchurchill")

        # We actually need the user name
        self.user = getpass.getuser()

        self.testDir = self.testInit.generateWorkDir()
        self.configFile = EmulatorSetup.setupWMAgentConfig()

    def tearDown(self):
        """
        Database deletion
        """
        self.testInit.clearDatabase(modules=["WMCore.WMBS", "WMCore.BossAir", "WMCore.ResourceControl"])
        self.testInit.delWorkDir()
        self.testInit.tearDownCouch()
        EmulatorSetup.deleteConfig(self.configFile)
        return

    def getConfig(self):
        """
        _getConfig_

        Build a basic JobTracker config
        """

        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        config.section_("Agent")
        config.Agent.agentName = 'testAgent'

        config.section_("CoreDatabase")
        config.CoreDatabase.connectUrl = os.getenv("DATABASE")
        config.CoreDatabase.socket = os.getenv("DBSOCK")

        # JobTracker
        config.component_("JobTracker")
        config.JobTracker.logLevel = 'INFO'
        config.JobTracker.pollInterval = 10
        config.JobTracker.trackerName = 'CondorTracker'
        config.JobTracker.pluginDir = 'WMComponent.JobTracker.Plugins'
        config.JobTracker.componentDir = os.path.join(os.getcwd(), 'Components')
        config.JobTracker.runTimeLimit = 7776000  # Jobs expire after 90 days
        config.JobTracker.idleTimeLimit = 7776000
        config.JobTracker.heldTimeLimit = 7776000
        config.JobTracker.unknTimeLimit = 7776000


        config.component_("JobSubmitter")
        config.JobSubmitter.logLevel = 'INFO'
        config.JobSubmitter.maxThreads = 1
        config.JobSubmitter.pollInterval = 10
        config.JobSubmitter.pluginName = 'AirPlugin'
        config.JobSubmitter.pluginDir = 'JobSubmitter.Plugins'
        config.JobSubmitter.submitDir = os.path.join(self.testDir, 'submit')
        config.JobSubmitter.submitNode = os.getenv("HOSTNAME", 'badtest.fnal.gov')
        # config.JobSubmitter.submitScript  = os.path.join(os.getcwd(), 'submit.sh')
        config.JobSubmitter.submitScript = os.path.join(WMCore.WMInit.getWMBASE(),
                                                         'test/python/WMComponent_t/JobSubmitter_t',
                                                         'submit.sh')
        config.JobSubmitter.componentDir = os.path.join(os.getcwd(), 'Components')
        config.JobSubmitter.workerThreads = 2
        config.JobSubmitter.jobsPerWorker = 200
        config.JobSubmitter.gLiteConf = os.path.join(os.getcwd(), 'config.cfg')



        # BossAir
        config.component_("BossAir")
        config.BossAir.pluginNames = ['TestPlugin', 'CondorPlugin']
        config.BossAir.pluginDir = 'WMCore.BossAir.Plugins'


        # JobStateMachine
        config.component_('JobStateMachine')
        config.JobStateMachine.couchurl = os.getenv('COUCHURL', 'cmssrv52.fnal.gov:5984')
        config.JobStateMachine.couchDBName = "jobtracker_t"

        return config




    def createTestJobs(self, nJobs, cacheDir):
        """
        _createTestJobs_

        Create several jobs
        """


        testWorkflow = Workflow(spec="spec.xml", owner="Simon",
                                name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset,
                                        workflow=testWorkflow,
                                        type="Processing",
                                        split_algo="FileBased")
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        # Create a file
        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(10, *[12312]))
        testFileA.setLocation('malpaquet')
        testFileA.create()

        baseName = makeUUID()

        # Now create a job
        for i in range(nJobs):
            testJob = Job(name='%s-%i' % (baseName, i))
            testJob.addFile(testFileA)
            testJob['location'] = 'malpaquet'
            testJob['retry_count'] = 1
            testJob['retry_max'] = 10
            testJob.create(testJobGroup)
            testJob.save()
            testJobGroup.add(testJob)

        testJobGroup.commit()

        # Set test job caches
        for job in testJobGroup.jobs:
            job.setCache(cacheDir)

        return testJobGroup



    @attr('integration')
    def testA_CondorTest(self):
        """
        _CondorTest_

        Because I don't want this test to be submitter dependent:
        Create a dummy condor job.
        Submit a dummy condor job.
        Track it.
        Kill it.
        Exit
        """

        # This has to be run with an empty queue
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        nJobs = 10
        jobCE = 'cmsosgce.fnal.gov/jobmanager-condor'

        # Create directories
        cacheDir = os.path.join(self.testDir, 'CacheDir')
        submitDir = os.path.join(self.testDir, 'SubmitDir')

        if not os.path.isdir(cacheDir):
            os.makedirs(cacheDir)
        if not os.path.isdir(submitDir):
            os.makedirs(submitDir)

        # Get config
        config = self.getConfig()

        # Get jobGroup
        testJobGroup = self.createTestJobs(nJobs=nJobs, cacheDir=cacheDir)

        # Propogate jobs
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')

        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nJobs)

        jobTracker = JobTrackerPoller(config)
        jobTracker.setup()



        # First iteration
        # There are no jobs in the tracker,
        # The tracker should register the jobs as missing
        # This should tell it that they've finished
        # So the tracker should send them onwards
        jobTracker.algorithm()

        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nJobs)


        result = self.getJobs.execute(state='complete', jobType="Processing")
        self.assertEqual(len(result), 0)



        # Second iteration
        # Reset the jobs
        # This time submit them to the queue
        # The jobs should remain in holding
        changer.propagate(testJobGroup.jobs, 'executing', 'created')

        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nJobs)

        # Create a submit script
        createSubmitScript(submitDir)


        jobPackage = os.path.join(self.testDir, 'JobPackage.pkl')
        f = open(jobPackage, 'w')
        f.write(' ')
        f.close()

        sandbox = os.path.join(self.testDir, 'sandbox.box')
        f = open(sandbox, 'w')
        f.write(' ')
        f.close()

        for job in testJobGroup.jobs:
            job['plugin'] = 'CondorPlugin'
            job['userdn'] = 'jchurchill'
            job['custom'] = {'location': 'malpaquet'}
            job['cache_dir'] = self.testDir
            job['sandbox'] = sandbox
            job['packageDir'] = self.testDir

        info = {}
        info['packageDir'] = self.testDir
        info['index'] = 0
        info['sandbox'] = sandbox

        jobTracker.bossAir.submit(jobs=testJobGroup.jobs, info=info)

        time.sleep(1)

        # All jobs should be running
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nJobs)


        # Run the algorithm.  After this
        # all jobs should still be running
        jobTracker.algorithm()

        # Are jobs in the right state?
        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nJobs)

        result = self.getJobs.execute(state='Complete', jobType="Processing")
        self.assertEqual(len(result), 0)

        # Are jobs still in the condor_q
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, nJobs)


        # Then we're done
        jobTracker.bossAir.kill(jobs=testJobGroup.jobs)

        # No jobs should be left
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0)

        jobTracker.algorithm()

        # Are jobs in the right state?
        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), 0)

        result = self.getJobs.execute(state='Complete', jobType="Processing")
        self.assertEqual(len(result), nJobs)


        # This is optional if you want to look at what
        # files were actually created during running
        # if os.path.isdir('testDir'):
        #    shutil.rmtree('testDir')
        # shutil.copytree('%s' %self.testDir, os.path.join(os.getcwd(), 'testDir'))


        return

    @attr('integration')
    def testB_ReallyLongTest(self):
        """
        _ReallyLongTest_

        Run a really long test using the condor plugin
        """

        # This has to be run with an empty queue
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        # This has to be run with an empty queue
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0, "User currently has %i running jobs.  Test will not continue" % (nRunning))

        nJobs = 500
        jobCE = 'cmsosgce.fnal.gov/jobmanager-condor'

        # Create directories
        cacheDir = os.path.join(self.testDir, 'CacheDir')
        submitDir = os.path.join(self.testDir, 'SubmitDir')

        if not os.path.isdir(cacheDir):
            os.makedirs(cacheDir)
        if not os.path.isdir(submitDir):
            os.makedirs(submitDir)

        # Get config
        config = self.getConfig()

        # Get jobGroup
        testJobGroup = self.createTestJobs(nJobs=nJobs, cacheDir=cacheDir)

        # Propogate jobs
        changer = ChangeState(config)
        changer.propagate(testJobGroup.jobs, 'created', 'new')
        changer.propagate(testJobGroup.jobs, 'executing', 'created')

        jobTracker = JobTrackerPoller(config)
        jobTracker.setup()

        # Now create some jobs
        for job in testJobGroup.jobs[:(nJobs // 2)]:
            jdl = createJDL(jobID=job['id'], directory=submitDir, jobCE=jobCE)
            jdlFile = os.path.join(submitDir, 'condorJDL_%i.jdl' % (job['id']))
            handle = open(jdlFile, 'w')
            handle.writelines(jdl)
            handle.close()

            command = ["condor_submit", jdlFile]
            pipe = subprocess.Popen(command, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, shell=False)
            pipe.communicate()

        startTime = time.time()
        cProfile.runctx("jobTracker.algorithm()", globals(), locals(), filename="testStats.stat")
        # jobTracker.algorithm()
        stopTime = time.time()

        # Are jobs in the right state?
        result = self.getJobs.execute(state='Executing', jobType="Processing")
        self.assertEqual(len(result), nJobs // 2)

        result = self.getJobs.execute(state='Complete', jobType="Processing")
        self.assertEqual(len(result), nJobs // 2)

        # Then we're done
        killList = [x['id'] for x in testJobGroup.jobs]
        jobTracker.killJobs(jobList=killList)

        # No jobs should be left
        nRunning = getCondorRunningJobs(self.user)
        self.assertEqual(nRunning, 0)

        print ("Process took %f seconds to process %i classAds" % ((stopTime - startTime),
                                                                  nJobs // 2))
        p = pstats.Stats('testStats.stat')
        p.sort_stats('cumulative')
        p.print_stats()


if __name__ == '__main__':
    unittest.main()



from Integration_t.RequestLifeCycleBase_t import RequestLifeCycleBase_t
import unittest

class TaskChainLifeCycle_t(unittest.TestCase, RequestLifeCycleBase_t):

    task1Config = {'TaskName' : 'DIGI', 'SplittingAlgorithm' : 'LumiBased',
                   'SplittingArguments' : {'lumis_per_job' : 8},
                   'InputDataset' : '/RelValSingleElectronPt10/CMSSW_6_0_0_pre3-START60_V0-v1/GEN-SIM',
                   'ConfigCacheID' : 'RelValSet_CMSSW_6_0_0_pre11_1_15.0DIGI'
    }

    task2Config = {'TaskName' : 'RECO', 'InputTask' : 'DIGI',
                   'SplittingAlgorithm' : 'LumiBased', 'SplittingArguments' : {'lumis_per_job' : 8},
                   'InputFromOutputModule' : 'FEVTDEBUGHLToutput',
                   'ConfigCacheID' : 'RelValSet_CMSSW_6_0_0_pre11_1_15.0RECO'



    }

    requestParams = {'CMSSWVersion' : 'CMSSW_4_4_2_patch2',
                        "Scenario": "pp", "inputMode" : "Scenario", "ProcScenario" : "pp",
                        "RequestType": "TaskChain", "TaskChain" : 2,
                        "Requestor": 'integration',
                        "GlobalTag": 'FT_R_44_V11::All', 'Group' : 'DMWM', "RequestPriority" : 10,
                        "Task1" : task1Config, "Task2" : task2Config,
                    }


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# encoding: utf-8
"""
CouchUtils_t.py

Created by Dave Evans on 2010-10-04.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from builtins import object

import unittest

import WMCore.Database.CouchUtils as CouchUtils
from WMQuality.TestInitCouchApp import TestInitCouchApp


class CouchUtils_t(unittest.TestCase):
    def setUp(self):
        self.testInit = TestInitCouchApp(__file__)
        self.testInit.setupCouch("wmcore-acdc-couchutils", "GroupUser", "ACDC")

    def tearDown(self):
        self.testInit.tearDownCouch()

    def testA(self):
        """object driven connection via initialiseCouch method"""

        class Thingy(object):
            """misc object with couch access attrs"""

            def __init__(self):
                self.couchdb = None
                self.database = None
                self.url = None

            @CouchUtils.connectToCouch
            def __call__(self):
                return True

        couchThingy = Thingy()

        # test throws with everything None
        self.assertRaises(CouchUtils.CouchConnectionError, CouchUtils.initialiseCouch, couchThingy)
        couchThingy.url = self.testInit.couchUrl
        self.assertRaises(CouchUtils.CouchConnectionError, CouchUtils.initialiseCouch, couchThingy)
        couchThingy.database = self.testInit.couchDbName

        try:
            CouchUtils.initialiseCouch(couchThingy)
        except Exception as ex:
            msg = "Error initialising couch client for test object:\n %s " % str(ex)
            self.fail(msg)

        self.assertIsNotNone(couchThingy.couchdb)
        # test decorator on already connected object
        try:
            couchThingy()
        except Exception as ex:
            msg = "Error invoking connectToCouch decorator:\n %s" % str(ex)
            self.fail(msg)

        newCouchThingy = Thingy()
        newCouchThingy.database = self.testInit.couchDbName
        newCouchThingy.url = self.testInit.couchUrl
        # 2nd call will make sure full connection is called
        try:
            newCouchThingy()
        except Exception as ex:
            msg = "Error invoking connectToCouch decorator:\n %s" % str(ex)
            self.fail(msg)
        self.assertIsNotNone(newCouchThingy)

    def testB(self):
        """check requirement tests"""

        class Thingy(dict):
            """test object with required attrs"""

            def __init__(self):
                super(Thingy, self).__init__()
                self.collection = "NotNone"
                self.owner = "NotNone"
                self['fileset_id'] = "NotNone"
                self['owner_id'] = "NotNone"

            @CouchUtils.requireCollection
            def call1(self):
                return True

            @CouchUtils.requireOwner
            def call4(self):
                return True

        thingy = Thingy()

        try:
            thingy.call1()
        except Exception as ex:
            msg = "Failure in requireCollection decorator: %s" % str(ex)
            self.fail(msg)
        try:
            thingy.call4()
        except Exception as ex:
            msg = "Failure in requireOwner decorator: %s" % str(ex)
            self.fail(msg)

        # now screw it up
        thingy.collection = None
        thingy.owner = None

        self.assertRaises(RuntimeError, thingy.call1)
        self.assertRaises(RuntimeError, thingy.call4)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
_InsertSpec_

Oracle implementation of DBSBuffer3.UpdateSpec

Created on Mar 11, 2013

@author: dballest
"""

from WMComponent.DBS3Buffer.MySQL.UpdateSpec import UpdateSpec as MySQLUpdateSpec

class UpdateSpec(MySQLUpdateSpec):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains a few basic WMCore-related utilitarian
Rucio data structures and functions
"""

import random

RUCIO_VALID_PROJECT = ("Production", "RelVal", "Tier0", "Test", "User")
# grouping values are extracted from:
# https://github.com/rucio/rucio/blob/master/lib/rucio/common/schema/cms.py#L117
GROUPING_DSET = "DATASET"
GROUPING_ALL = "ALL"
RUCIO_RULES_PRIORITY = {"low": 2, "normal": 3, "high": 4, "reserved": 5}
# number of copies to be defined when creating replication rules
NUM_COPIES_DEFAULT = 1
NUM_COPIES_NANO = 2

def validateMetaData(did, metaDict, logger):
    """
    This function can be extended in the future, for now it will only
    validate the DID creation metadata, more specifically only the
    "project" parameter
    :param did: the DID that will be inserted
    :param metaDict: a dictionary with all the DID metadata data to be inserted
    :param logger: a logger object
    :return: False if validation fails, otherwise True
    """
    if metaDict.get("project", "Production") in RUCIO_VALID_PROJECT:
        return True
    msg = f"DID: {did} has an invalid 'project' meta-data value: {metaDict['project']}"
    msg += f"The supported 'project' values are: {str(RUCIO_VALID_PROJECT)}"
    logger.error(msg)
    return False


def weightedChoice(rses, rseWeights):
    """
    Given a list of items and their respective weights (quota in this case),
    perform a weighted selection.
    :param rses: a list of tuples with the RSE name and whether it requires approval
    :param rseWeights: a list with RSE weights (quota)
    :return: a tuple from the choices list
    """
    listChoice = random.choices(population=rses, weights=rseWeights, k=1)
    # return only the tuple, not the list with a tuple item
    return listChoice[0]


def isTapeRSE(rseName):
    """
    Given an RSE name, return True if it's a Tape RSE (rse_type=TAPE), otherwise False
    :param rseName: string with the RSE name
    :return: True or False
    """
    # NOTE: a more reliable - but more expensive - way to know that would be
    # to query `get_rse` and evaluate the rse_type parameter
    return rseName.endswith("_Tape")


def dropTapeRSEs(listRSEs):
    """
    Method to parse a list of RSE names and return only those that
    are not a rse_type=TAPE, so in general only Disk endpoints
    :param listRSEs: list with the RSE names
    :return: a new list with only DISK RSE names
    """
    diskRSEs = []
    for rse in listRSEs:
        if isTapeRSE(rse):
            continue
        diskRSEs.append(rse)
    return diskRSEs



#!/usr/bin/env python
# encoding: utf-8
"""
Locale.py

Template for a locale configuration that contains all the stuff that agent operators want to customise
on a regular basis

Created by Dave Evans on 2011-05-20.
Copyright (c) 2011 Fermilab. All rights reserved.
"""

import sys
import os
import socket

from WMCore.Configuration import Configuration

config = Configuration()
config.section_('locale')

#
# Define the required sections
#
config.locale.section_("reqmgr")
config.locale.section_("gwq")
config.locale.section_("lwq")
config.locale.section_("agent")
config.locale.section_("crabserver")
config.locale.section_("couch")
config.locale.section_("mysql")
config.locale.section_("certificates")

# canned settings for each section
# request manager
config.locale.reqmgr.hostname = "cmssrv98.fnal.gov"
config.locale.reqmgr.port = "9798"
config.locale.reqmgr.database = "reqmgrdb"

# global workqueue
config.locale.gwq.hostname = "cmssrv98.fnal.gov"
config.locale.gwq.port = "9996"
config.locale.gwq.database = "workqueuedb"

# local workqueue
config.locale.lwq.port = "9997"
config.locale.lwq.database = "workqueuedb"

#crabserver
config.locale.crabserver.host = "cmssrv98.fnal.gov"
config.locale.crabserver.port = "9600"

# couch installation
config.locale.couch.port = "5984"
config.locale.couch.host = "0.0.0.0"
config.locale.couch.hostname = socket.gethostname()
config.locale.couch.url = None

#mysql settings
config.locale.mysql.socket = None
config.locale.mysql.url = None

# agent settings
config.locale.agent.teamName = "DMWM"
config.locale.agent.name = os.environ['USER']
config.locale.agent.hostname = socket.gethostname()
config.locale.agent.database = "wmagentdb"

# make it easy to import the Locale config as Locale:
Locale = config



#! /bin/env python

"""
A service class for retrieving data from McM using
an SSO cookie since it sits behind CERN SSO
"""

from builtins import str, object
import json
import os
import pycurl
import subprocess
import logging
from typing import List
from io import BytesIO
from WMCore.WMException import WMException


class McMNoDataError(WMException):
    """
    _McMNoDataError_
        McM responded but has no data for the request
    """

    def __init__(self):
        WMException.__init__(self, 'McM responded correctly but has no data')


class McM(object):
    """
    A service class for retrieving data from McM using
    an SSO cookie since it sits behind CERN SSO
    'key' must be unencrypted
    """

    def __init__(self, url='https://cms-pdmv.cern.ch/mcm', tmpDir='/tmp'):
        self.url = url
        self.tmpDir = tmpDir
        self.logger = self._get_logger()
        self.cookieFile = None

    def __enter__(self):
        self._krb_ticket()
        self._get_cookie()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        if self.cookieFile:
            try:
                os.remove(self.cookieFile)
            except OSError:
                return
        return

    def _get_logger(self) -> logging.Logger:
        """
        Create a logger for McM client
        """
        logger: logging.Logger = logging.getLogger("mcm_client")
        date_format: str = "%Y-%m-%d %H:%M:%S %z"
        format: str = "[%(levelname)s][%(name)s][%(asctime)s]: %(message)s"
        formatter: logging.Formatter = logging.Formatter(fmt=format, datefmt=date_format)
        handler: logging.StreamHandler = logging.StreamHandler()

        handler.setLevel(logging.INFO)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        return logger

    def _krb_ticket(self) -> None:
        """
        Check there is a valid Kerberos ticket for requesting a SSO
        cookie. Raise a RuntimeError if there is not one.
        """
        process = subprocess.Popen(["klist", "-f"],
                                   stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=False)
        stdout: str = process.communicate()[0].decode("utf-8")
        if process.returncode != 0:
            msg: str = ("There is no valid Kerberos ticket for requesting a SSO cookie. "
                        "Please make sure to provide one for your runtime environment"
                        )
            self.logger.error(msg)
            raise RuntimeError("FATAL -- %s\nError msg: %s" % (msg, stdout))

    def _get_cookie(self) -> None:
        """
        Request a SSO cookie to authenticate to McM.
        """
        self.cookieFile = os.path.join(self.tmpDir, 'ssoCookie.txt')
        command: List[str] = ["auth-get-sso-cookie", "-u", self.url, "-o", self.cookieFile, "-vv"]
        callback_not_invoked: str = "DEBUG: Not automatically redirected: trying SAML authentication"
        cookie_stored: str = "INFO: Saving cookies"

        process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT, shell=False)
        stdout: str = process.communicate()[0].decode("utf-8")
        stdout_list: List[str] = stdout.strip().split("\n")
        if process.returncode != 0:
            raise RuntimeError("FATAL -- Error requesting SSO cookie\nError msg: %s" % (stdout))

        if callback_not_invoked in stdout:
            callback_idx: int = stdout_list.index(callback_not_invoked)
            stored_after_issue: bool = cookie_stored in stdout_list[callback_idx + 1]
            if stored_after_issue:
                msg: str = ("Callback method was not invoked. "
                            "Please make sure the provided Kerberos ticket is not linked to an account with 2FA"
                            )
                raise RuntimeError(msg)

    def _getURL(self, extendURL):
        """
        Fetch an MCM URL with CURL using the SSO cookie.
        Only intended to be used internally
        """

        try:
            b = BytesIO()
            c = pycurl.Curl()

            fullUrl = '%s/%s' % (self.url, extendURL)

            c.setopt(c.URL, fullUrl)
            c.setopt(c.SSL_VERIFYPEER, False)
            c.setopt(c.SSL_VERIFYHOST, False)
            c.setopt(c.FOLLOWLOCATION, True)
            c.setopt(c.COOKIEJAR, self.cookieFile)
            c.setopt(c.COOKIEFILE, self.cookieFile)
            c.setopt(c.WRITEFUNCTION, b.write)
            c.perform()
            if c.getinfo(pycurl.HTTP_CODE) != 200:
                raise IOError
        except:
            c.close()
            raise IOError('Was not able to fetch or decode URL from McM')

        try:
            body = b.getvalue()
            res = json.loads(body)
        except ValueError:
            c.close()
            raise IOError('Was not able to decode JSON from McM')

        c.close()
        return res

    def getHistory(self, prepID):
        """
        Get the history record which has who did what to an McM request
        """

        try:
            url = 'search?db_name=batches&contains=%s&get_raw' % prepID
            res = self._getURL(url)
            history = res.get("results", [])[0].get("history", [])
            return history
        except IndexError:
            raise McMNoDataError

    def getRequest(self, prepID):
        """
        Get the request record which has, among other things,
        the number of requested events
        """

        url = 'public/restapi/requests/get/%s' % prepID
        res = self._getURL(url)
        return res['results']


if __name__ == '__main__':
    with McM() as mcm:
        history = mcm.getHistory(prepID='BTV-Upg2023SHCAL14DR-00002')
        request = mcm.getRequest(prepID='BTV-Upg2023SHCAL14DR-00002')



from __future__ import print_function

from builtins import str, bytes, int
from future.utils import viewitems
from Utils.PythonVersion import PY2

import sys
import types


class _EmptyClass(object):
    pass


class JSONThunker(object):
    """
    _JSONThunker_
    Converts an arbitrary object to <-> from a jsonable object.

    Will, for the most part "do the right thing" about various instance objects
    by storing their class information along with their data in a dict. Handles
    a recursion limit to prevent infinite recursion.

    self.passThroughTypes - stores a list of types that should be passed
      through unchanged to the JSON parser

    self.blackListedModules - a list of modules that should not be stored in
      the JSON.

    """

    def __init__(self):
        self.passThroughTypes = (type(None),
                                 bool,
                                 int,
                                 float,
                                 complex,
                                 str,
                                 bytes,
                                 )
        # objects that inherit from dict should be treated as a dict
        #   they don't store their data in __dict__. There was enough
        #   of those classes that it warrented making a special case
        self.dictSortOfObjects = (('WMCore.Datastructs.Job', 'Job'),
                                  ('WMCore.WMBS.Job', 'Job'),
                                  ('WMCore.Database.CMSCouch', 'Document'))
        # ditto above, but for lists
        self.listSortOfObjects = (('WMCore.DataStructs.JobPackage', 'JobPackage'),
                                  ('WMCore.WMBS.JobPackage', 'JobPackage'),)

        self.foundIDs = {}
        # modules we don't want JSONed
        self.blackListedModules = ('sqlalchemy.engine.threadlocal',
                                   'WMCore.Database.DBCore',
                                   'logging',
                                   'WMCore.DAOFactory',
                                   'WMCore.WMFactory',
                                   'WMFactory',
                                   'WMCore.Configuration',
                                   'WMCore.Database.Transaction',
                                   'threading',
                                   'datetime')

    def checkRecursion(self, data):
        """
        handles checking for infinite recursion
        """
        if id(data) in self.foundIDs:
            if self.foundIDs[id(data)] > 5:
                self.unrecurse(data)
                return "**RECURSION**"
            else:
                self.foundIDs[id(data)] += 1
                return data
        else:
            self.foundIDs[id(data)] = 1
            return data

    def unrecurse(self, data):
        """
        backs off the recursion counter if we're returning from _thunk
        """
        try:
            self.foundIDs[id(data)] -= 1
        except:
            print("Could not find count for id %s of type %s data %s" % (id(data), type(data), data))
            raise

    def checkBlackListed(self, data):
        """
        checks to see if a given object is from a blacklisted module
        """
        try:
            # special case
            if data.__class__.__module__ == 'WMCore.Database.CMSCouch' and data.__class__.__name__ == 'Document':
                data.__class__ = type({})
                return data
            if data.__class__.__module__ in self.blackListedModules:
                return "Blacklisted JSON object: module %s, name %s, str() %s" % \
                       (data.__class__.__module__, data.__class__.__name__, str(data))
            else:
                return data
        except Exception:
            return data

    def thunk(self, toThunk):
        """
        Thunk - turns an arbitrary object into a JSONable object
        """
        self.foundIDs = {}
        data = self._thunk(toThunk)
        return data

    def unthunk(self, data):
        """
        unthunk - turns a previously 'thunked' object back into a python object
        """
        return self._unthunk(data)

    def handleSetThunk(self, toThunk):
        toThunk = self.checkRecursion(toThunk)
        tempDict = {'thunker_encoded_json': True, 'type': 'set'}
        tempDict['set'] = self._thunk(list(toThunk))
        self.unrecurse(toThunk)
        return tempDict

    def handleListThunk(self, toThunk):
        toThunk = self.checkRecursion(toThunk)
        for k, v in enumerate(toThunk):
            toThunk[k] = self._thunk(v)
        self.unrecurse(toThunk)
        return toThunk

    def handleDictThunk(self, toThunk):
        toThunk = self.checkRecursion(toThunk)
        special = False
        tmpdict = {}
        for k, v in viewitems(toThunk):
            if type(k) == type(int):
                special = True
                tmpdict['_i:%s' % k] = self._thunk(v)
            elif type(k) == type(float):
                special = True
                tmpdict['_f:%s' % k] = self._thunk(v)
            else:
                tmpdict[k] = self._thunk(v)
        if special:
            toThunk['thunker_encoded_json'] = self._thunk(True)
            toThunk['type'] = self._thunk('dict')
            toThunk['dict'] = tmpdict
        else:
            toThunk.update(tmpdict)
        self.unrecurse(toThunk)
        return toThunk

    def handleObjectThunk(self, toThunk):
        toThunk = self.checkRecursion(toThunk)
        toThunk = self.checkBlackListed(toThunk)

        if isinstance(toThunk, (str, bytes)):
            # things that got blacklisted
            return toThunk
        if hasattr(toThunk, '__to_json__'):
            # Use classes own json thunker
            toThunk2 = toThunk.__to_json__(self)
            self.unrecurse(toThunk)
            return toThunk2
        elif isinstance(toThunk, dict):
            toThunk2 = self.handleDictObjectThunk(toThunk)
            self.unrecurse(toThunk)
            return toThunk2
        elif isinstance(toThunk, list):
            # a mother thunking list
            toThunk2 = self.handleListObjectThunk(toThunk)
            self.unrecurse(toThunk)
            return toThunk2
        else:
            try:
                thunktype = '%s.%s' % (toThunk.__class__.__module__,
                                       toThunk.__class__.__name__)
                tempDict = {'thunker_encoded_json': True, 'type': thunktype}
                tempDict[thunktype] = self._thunk(toThunk.__dict__)
                self.unrecurse(toThunk)
                return tempDict
            except Exception as e:
                tempDict = {'json_thunk_exception_': "%s" % e}
                self.unrecurse(toThunk)
                return tempDict

    def handleDictObjectThunk(self, data):
        thunktype = '%s.%s' % (data.__class__.__module__,
                               data.__class__.__name__)
        tempDict = {'thunker_encoded_json': True,
                    'is_dict': True,
                    'type': thunktype,
                    thunktype: {}}

        for k, v in viewitems(data.__dict__):
            tempDict[k] = self._thunk(v)
        for k, v in viewitems(data):
            tempDict[thunktype][k] = self._thunk(v)

        return tempDict

    def handleDictObjectUnThunk(self, value, data):
        data.pop('thunker_encoded_json', False)
        data.pop('is_dict', False)
        thunktype = data.pop('type', False)

        for k, v in viewitems(data):
            if k == thunktype:
                for k2, v2 in viewitems(data[thunktype]):
                    value[k2] = self._unthunk(v2)
            else:
                value.__dict__[k] = self._unthunk(v)
        return value

    def handleListObjectThunk(self, data):
        thunktype = '%s.%s' % (data.__class__.__module__,
                               data.__class__.__name__)
        tempDict = {'thunker_encoded_json': True,
                    'is_list': True,
                    'type': thunktype,
                    thunktype: []}
        for k, v in enumerate(data):
            tempDict['thunktype'].append(self._thunk(v))
        for k, v in viewitems(data.__dict__):
            tempDict[k] = self._thunk(v)
        return tempDict

    def handleListObjectUnThunk(self, value, data):
        data.pop('thunker_encoded_json', False)
        data.pop('is_list', False)
        thunktype = data.pop('type')
        for k, v in viewitems(data[thunktype]):
            setattr(value, k, self._unthunk(v))

        for k, v in viewitems(data):
            if k == thunktype:
                continue
            value.__dict__ = self._unthunk(v)
        return value

    def _thunk(self, toThunk):
        """
        helper function for thunk, does the actual work
        """
        if isinstance(toThunk, self.passThroughTypes):
            return toThunk
        elif type(toThunk) is list:
            return self.handleListThunk(toThunk)
        elif type(toThunk) is dict:
            return self.handleDictThunk(toThunk)
        elif type(toThunk) is set:
            return self.handleSetThunk(toThunk)
        elif type(toThunk) is types.FunctionType:
            self.unrecurse(toThunk)
            return "function reference"
        elif isinstance(toThunk, object):
            return self.handleObjectThunk(toThunk)
        else:
            self.unrecurse(toThunk)
            raise RuntimeError(type(toThunk))

    def _unthunk(self, jsondata):
        """
        _unthunk - does the actual work for unthunk
        """
        if PY2 and type(jsondata) is str:
            return jsondata.encode("utf-8")
        if type(jsondata) is dict:
            if 'thunker_encoded_json' in jsondata:
                # we've got a live one...
                if jsondata['type'] == 'set':
                    newSet = set()
                    for i in self._unthunk(jsondata['set']):
                        newSet.add(self._unthunk(i))
                    return newSet
                if jsondata['type'] == 'dict':
                    # We have a "special" dict
                    data = {}
                    for k, v in viewitems(jsondata['dict']):
                        tmp = self._unthunk(v)
                        if k.startswith('_i:'):
                            data[int(k.lstrip('_i:'))] = tmp
                        elif k.startswith('_f:'):
                            data[float(k.lstrip('_f:'))] = tmp
                        else:
                            data[k] = tmp
                    return data
                else:
                    # spawn up an instance.. good luck
                    #   here be monsters
                    #   inspired from python's pickle code
                    ourClass = self.getThunkedClass(jsondata)

                    value = _EmptyClass()
                    if hasattr(ourClass, '__from_json__'):
                        # Use classes own json loader
                        try:
                            value.__class__ = ourClass
                        except Exception:
                            value = ourClass()
                        value = ourClass.__from_json__(value, jsondata, self)
                    elif 'thunker_encoded_json' in jsondata and 'is_dict' in jsondata:
                        try:
                            value.__class__ = ourClass
                        except Exception:
                            value = ourClass()
                        value = self.handleDictObjectUnThunk(value, jsondata)
                    elif 'thunker_encoded_json' in jsondata:
                        try:
                            value.__class__ = ourClass
                        except Exception:
                            value = ourClass()
                        value = self.handleListObjectUnThunk(value, jsondata)
                    else:
                        raise RuntimeError('Could not unthunk a class. Code to try was removed because it had errors.')
                return value
            else:
                data = {}
                for k, v in viewitems(jsondata):
                    data[k] = self._unthunk(v)
                return data

        else:
            return jsondata

    @staticmethod
    def getThunkedClass(jsondata):
        """
        Work out the class from it's thunked json representation
        """
        module = jsondata['type'].rsplit('.', 1)[0]
        name = jsondata['type'].rsplit('.', 1)[1]
        if (module == 'WMCore.Services.Requests') and (name == JSONThunker):
            raise RuntimeError("Attempted to unthunk a JSONThunker..")

        __import__(module)
        mod = sys.modules[module]
        ourClass = getattr(mod, name)
        return ourClass



"""
_GetCompletedBlocks_

MySQL implementation of RucioInjector.GetClompletedBlocks

Retrieve a list of blocks that are complete,
including their location and the sites they
are subscribed too

"""


from WMCore.Database.DBFormatter import DBFormatter


class GetCompletedBlocks(DBFormatter):
    """
    Retrieves a list of blocks that are closed but NOT sure yet if they are deleteable:
      - The workflows for all files in the block need to be completed (This relates only to the
        workflows directly producing the files and does not track child workflows completion)
      - The subscription made for dataset is copy+delete
      - A subscription has been made at the Data Management system
      - The blocks hasn't been deleted yet
    """
    sql = """SELECT dbsbuffer_block.blockname,
                    dbsbuffer_location.pnn,
                    dbsbuffer_dataset.path,
                    dbsbuffer_dataset_subscription.site,
                    dbsbuffer_workflow.name,
                    dbsbuffer_block.create_time
             FROM dbsbuffer_dataset_subscription
             INNER JOIN dbsbuffer_dataset ON
               dbsbuffer_dataset.id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_block ON
               dbsbuffer_block.dataset_id = dbsbuffer_dataset_subscription.dataset_id
             INNER JOIN dbsbuffer_file ON
               dbsbuffer_file.block_id = dbsbuffer_block.id
             INNER JOIN dbsbuffer_workflow ON
               dbsbuffer_workflow.id = dbsbuffer_file.workflow
             INNER JOIN dbsbuffer_location ON
               dbsbuffer_location.id = dbsbuffer_block.location
             WHERE dbsbuffer_dataset_subscription.delete_blocks = 1
               AND dbsbuffer_dataset_subscription.subscribed = 1
               AND dbsbuffer_block.status = 'Closed'
               AND dbsbuffer_block.deleted = 0
             GROUP BY dbsbuffer_block.blockname,
                      dbsbuffer_location.pnn,
                      dbsbuffer_dataset.path,
                      dbsbuffer_dataset_subscription.site,
                      dbsbuffer_workflow.name,
                      dbsbuffer_block.create_time
             """

    def format(self, result):
        """
        _format_

        Format the query results into the proper dictionary expected at the upper layer Python code.
        The input should be a list of database objects each representing a line returned from the database
        with key names matching the column names from the sql query
        The result should be a list of dictionaries one record per block returned from the database
        with key names mapped to the python code expected structures. All workflows and sites are aggregated 
        into the same block record.

        e.g.
        { '/Tau/Run2022C-PromptReco-v1/MINIAOD#2dd5a82b-873a-4403-8da1-6b943dac7081': {'blockCreateTime': 1659675842,
                                                                              'blockName': '/Tau/Run2022C-PromptReco-v1/MINIAOD#2dd5a82b-873a-4403-8da1-6b943dac7081',
                                                                              'dataset': '/Tau/Run2022C-PromptReco-v1/MINIAOD',
                                                                              'location': 'T0_CH_CERN_Disk',
                                                                              'sites': {'T1_ES_PIC_Disk',
                                                                                        'T1_ES_PIC_MSS'},
                                                                              'workflowNames': {'PromptReco_Run356614_Tau'}},
          '/Tau/Run2022C-PromptReco-v1/MINIAOD#f6bf5cc7-cab2-4572-8f30-574296bb109d': {'blockCreateTime': 1659723755,
                                                                              'blockName': '/Tau/Run2022C-PromptReco-v1/MINIAOD#f6bf5cc7-cab2-4572-8f30-574296bb109d',
                                                                              'dataset': '/Tau/Run2022C-PromptReco-v1/MINIAOD',
                                                                              'location': 'T0_CH_CERN_Disk',
                                                                              'sites': {'T1_ES_PIC_Disk',
                                                                                        'T1_ES_PIC_MSS'},
                                                                              'workflowNames': {'PromptReco_Run356615_Tau',
                                                                                                'PromptReco_Run356619_Tau'}}
        }


        NOTE:
         * location: Means where the output block has been created
         * site(s):  Means where the dataset gets a container-level rule

        :param result: The result as returned by the mysql query execution.
        :return:       Dictionary of dictionaries, each one describing a block.
        """

        # NOTE: We need to rename all the keys to follow the cammelCase standard. And also to comply
        #       with the key names as expected from the rest of the already existing python code
        keyMap = {'blockname': 'blockName',
                  'name': 'workflowNames',
                  'pnn': 'location',
                  'site': 'sites',
                  'path': 'dataset',
                  'create_time': 'blockCreateTime'}

        listResults = DBFormatter.formatDict(self, result)
        dictResults = {}
        for record in listResults:
            # Populates results dict and adds all workflows and sites of the same block to a single record
            blockName = record['blockname']
            if blockName in dictResults:
                dictResults[blockName]['workflowNames'].add(record['name'])
                dictResults[blockName]['sites'].add(record['site'])
            else:
                for dbKey, pyKey in keyMap.items():
                    if dbKey == 'site' or dbKey == 'name':
                        data = record.pop(dbKey)
                        record[pyKey] = set()
                        record[pyKey].add(data)
                    else:
                        record[pyKey] = record.pop(dbKey)
                dictResults[blockName] = record

        return dictResults

    def execute(self, conn=None, transaction=False, returnCursor=False):
        """
        Executing the current sql query.
        :param conn:        A current database connection to be used if existing
        :param transaction: A current database transaction to be used if existing
        :return:            A list of dictionaries one record for each database line returned
        """
        results = self.dbi.processData(self.sql, conn=conn,
                                       transaction=transaction)

        return self.format(results)



#!/usr/bin/env python
"""
_REST_t_

Unit tests for checking RESTModel works correctly

TODO: duplicate all direct call tests to ones that use HTTP
"""

import json
import unittest
import logging

from nose.plugins.attrib import attr

from WMCore.WebTools.RESTFormatter import RESTFormatter
from WMQuality.WebTools.RESTBaseUnitTest import RESTBaseUnitTest
from WMQuality.WebTools.RESTClientAPI import methodTest
from WMQuality.WebTools.RESTServerSetup import DefaultConfig

from Utils.PythonVersion import PY3

class RESTFormatTest(RESTBaseUnitTest):

    def initialize(self):
        self.config = DefaultConfig('WMCore_t.WebTools_t.DummyRESTModel')
        do_debug = False

        self.config.Webtools.environment = 'development'
        if do_debug:
            self.config.Webtools.error_log_level = logging.DEBUG
            self.config.Webtools.access_log_level = logging.DEBUG
        else:
            self.config.Webtools.error_log_level = logging.WARNING
            self.config.Webtools.access_log_level = logging.WARNING

        self.urlbase = self.config.getServerUrl()

    def testUnsupportedFormat(self):
        # test not accepted type should return 406 error
        url = self.urlbase +'list1/'
        methodTest('GET', url, accept='text/das', output={'code':406})

    def testSupportedFormat(self):
        rf = RESTFormatter(config=self.config.Webtools)
        url = self.urlbase +'list1/'

        for textType in rf.supporttypes.keys():
            # test accepted type should return 200 error
            methodTest('GET', url, accept=textType, output={'code':200})

    # This test is flipping back and forth in Jenkins. Perhaps due to port 8888 not being available.
    # Disabling for now
    @attr("integration")
    def testEncodedInput(self):
        textType = 'text/plain'

        url = self.urlbase + 'list3?a=a%&b=b'
        data = json.dumps({'a': 'a%', 'b': 'b'})
        methodTest('GET', url, accept=textType,
                         output={'code':200, 'data':data})

        request_input={'a':'%', 'b':'b'}

        #methodTest encoded input with urlencode
        url = self.urlbase +'list3'
        data = json.dumps({'a': '%', 'b': 'b'})
        methodTest('GET', url, accept=textType, request_input=request_input,
                 output={'code':200, 'data':data})

    def testReturnFormat(self):
        return_type = 'application/json'

        url = self.urlbase +'list3?a=a%&b=b'
        methodTest('GET', url, accept=return_type,
                         output={'code':200, 'data':'{"a": "a%", "b": "b"}'})

        url = self.urlbase + 'list?input_int=a&input_str=a'
        expected_data_py2 = '{"exception": 400, "message": "Invalid input: Input data failed validation.", "type": "HTTPError"}'
        expected_data_py3 = '{"exception": 400, "type": "HTTPError", "message": "Invalid input: Input data failed validation."}'
        expected_data = expected_data_py3 if PY3 else expected_data_py2
        methodTest('GET', url, accept=return_type,
                         output={'code':400, 'data':expected_data})

    def testNoArgMethods(self):
        """
        list1 takes no arguments, it should raise an error if called with one. Require json output.
        """
        return_type = 'application/json'
        url = self.urlbase + 'list1?int=a'
        expected_data_py2 = '{"exception": 400, "message": "Invalid input: Arguments added where none allowed", "type": "HTTPError"}'
        expected_data_py3 = '{"exception": 400, "type": "HTTPError", "message": "Invalid input: Arguments added where none allowed"}'
        expected_data = expected_data_py3 if PY3 else expected_data_py2
        methodTest('GET', url, accept=return_type, output={'code':400, 'data':expected_data})

    def testGenerator(self):
        rf = RESTFormatter(config=self.config.Webtools)
        url = self.urlbase +'gen'
        # gen method from DummyRESTModel will return this generator
        gen = ({'idx':i} for i in range(10))
        # the WMCore should convert it into list regardless of accept type
        data = rf.json(gen)
        methodTest('GET', url, accept='application/json',
                         output={'code':200, 'data':data})
        methodTest('GET', url, accept='*/*',
                         output={'code':200, 'data':data})

if __name__ == "__main__":
    unittest.main()



"""
MockReqMgrAux class provides mocking methods of ReqMgrAux class
"""
# futures
from __future__ import (division, print_function)
from builtins import object

# system modules
import time

# WMCore modues
from WMCore.Agent.DefaultConfig import DEFAULT_AGENT_CONFIG


class MockReqMgrAux(object):
    """
    Version of Services/PhEDEx intended to be used with mock or unittest.mock
    """

    def __init__(self, *args, **kwargs):
        print("Using MockReqMgrAux", args, kwargs)
        tstamp = int(time.time())
        cname = 'testCampaign'
        cdict = {"CampaignName": cname,
                 "PrimaryAAA": True,
                 "SecondaryAAA": False,
                 "SiteWhiteList": ["T1", "T2"],
                 "SiteBlackList": [],
                 "SecondaryLocation": ["T1", "T2"],
                 "Secondaries": {"datasetA": ["T1", "T2"], "datasetB": ["T1"]},
                 "MaxCopies": 2,
                 "TiersToDM": [],
                 "PartialCopy": 1
                 }
        self.campaigns = [cdict]
        self.transferRecords = []
        rec = {"dataset": "/a/b/c", "dataType": "primary",
               "transferIDs": [1, 2], "completion": [1], "campaignName": cname}
        self.transferRecords = [{'workflowName': 'test', 'lastUpdate': tstamp, 'transfers': [rec]}]

    def getWMAgentConfig(self, agentName):
        """
        macking getWMAgentConfig returns default config.
        """
        return DEFAULT_AGENT_CONFIG

    def getCampaignConfig(self, name):
        "mocking getCampaignConfig method"
        campaigns = []
        if name == 'ALL_DOCS':
            return self.campaigns
        for camp in self.campaigns:
            if camp['CampaignName'] == name:
                campaigns.append(camp)
        return campaigns

    def getTransferInfo(self, name):
        "mocking getTransferInfo method"
        if name == 'ALL_DOCS':
            return self.transferRecords
        records = []
        for rec in self.transferRecords:
            if rec['workflowName'] == name:
                records.append(rec)
        return records

    def updateTransferInfo(self, name, doc, inPlace=False):
        "mocking updateTransferInfo method"
        # while mocking we do nothing
        print('mock method updateTransferInfo: name=%s, doc=%s inPlace=%s' \
              % (name, doc, inPlace))



from __future__ import (division, print_function)

import traceback
from WMCore.REST.CherryPyPeriodicTask import CherryPyPeriodicTask
from WMCore.Services.WMStats.WMStatsWriter import WMStatsWriter

class CleanUpTask(CherryPyPeriodicTask):
    """
    This class is used for both T0WMStats and WMStats
    controlled by config.reqdb_couch_app value
    """

    def __init__(self, rest, config):

        super(CleanUpTask, self).__init__(config)
        self.wmstatsDB = WMStatsWriter(config.wmstats_url, reqdbURL=config.reqmgrdb_url,
                              reqdbCouchApp=config.reqdb_couch_app)

    def setConcurrentTasks(self, config):
        """
        sets the list of functions which runs concurrently
        """
        self.concurrentTasks = [{'func': self.cleanUpArchivedRequests, 'duration': config.archivedCleanUpDuration}]

    def cleanUpArchivedRequests(self, config):
        """
        loop through the workflows in couchdb, if archived delete all the data in couchdb
        """
        self.logger.info("getting archived data")
        requestNames = self.wmstatsDB.getArchivedRequests()
        self.logger.info("archived list %s", requestNames)

        for req in requestNames:
            self.logger.info("Deleting data for: %s", req)
            try:
                result = self.wmstatsDB.deleteDocsByWorkflow(req)
            except Exception as ex:
                self.logger.error("deleting %s failed: %s", req, str(ex))
                for line in traceback.format_exc().rstrip().split("\n"):
                    self.logger.error(" " + line)
            else:
                if result is None:
                    self.logger.info("there were no documents to delete.")
                else:
                    self.logger.info("%s docs deleted", len(result))
        return



#!/usr/bin/env python
"""
_SetLastUpdate_

Oracle implementation of SetLastUpdateFileset

"""
__all__ = []



from WMCore.WMBS.MySQL.Fileset.SetLastUpdate import SetLastUpdate as SetLastUpdateFilesetMySQL

class SetLastUpdate(SetLastUpdateFilesetMySQL):
    sql = SetLastUpdateFilesetMySQL.sql



"""
_ProcessPool_t

Unit tests for the ProcessPool class.
"""

from builtins import range
import unittest
import nose

from WMCore.ProcessPool.ProcessPool import ProcessPool
from WMQuality.TestInit import TestInit

class ProcessPoolTest(unittest.TestCase):
    def setUp(self):
        """
        _setUp_

        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection(destroyAllDatabase = True)
        self.testInit.setSchema(customModules = ["WMCore.Agent.Database"],
                                useDefault = False)
        return

    def tearDown(self):
        """
        _tearDown_

        """
        self.testInit.clearDatabase()
        return

    def testA_ProcessPool(self):
        """
        _testProcessPool_

        """
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        config.Agent.useHeartbeat = False
        self.testInit.generateWorkDir(config)

        processPool = ProcessPool("ProcessPool_t.ProcessPoolTestWorker",
                                  totalSlaves = 1,
                                  componentDir = config.General.workDir,
                                  config = config,
                                  namespace = "WMCore_t")

        processPool.enqueue(["One", "Two", "Three"])
        result =  processPool.dequeue(3)

        self.assertEqual(len(result), 3, "Error: Expected three items back.")
        self.assertTrue( "One" in result)
        self.assertTrue( "Two" in result)
        self.assertTrue( "Three" in result)

        return

    def testB_ProcessPoolStress(self):
        """
        _testProcessPoolStress_

        """
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        config.Agent.useHeartbeat = False
        self.testInit.generateWorkDir(config)

        processPool = ProcessPool("ProcessPool_t.ProcessPoolTestWorker",
                                  totalSlaves = 1,
                                  componentDir = config.General.workDir,
                                  namespace = "WMCore_t",
                                  config = config)

        result = None
        input  = None
        for i in range(1000):
            input = []
            while i > 0:
                input.append("COMMAND%s" % i)
                i -= 1

            processPool.enqueue(input)
            result =  processPool.dequeue(len(input))

            self.assertEqual(len(result), len(input),
                             "Error: Wrong number of results returned.")

        for k in result:
            self.assertTrue(k in input)

        return


    def testC_MultiPool(self):
        """
        _testMultiPool_

        Run a test with multiple workers
        """
        raise nose.SkipTest
        config = self.testInit.getConfiguration()
        config.Agent.useHeartbeat = False
        self.testInit.generateWorkDir(config)

        processPool = ProcessPool("ProcessPool_t.ProcessPoolTestWorker",
                                  totalSlaves = 3,
                                  componentDir = config.General.workDir,
                                  namespace = "WMCore_t",
                                  config = config)

        for i in range(100):
            input = []
            while i > 0:
                input.append("COMMAND%s" % i)
                i -= 1

            processPool.enqueue(input)
            result =  processPool.dequeue(len(input))

            self.assertEqual(len(result), len(input),
                             "Error: Wrong number of results returned.")





if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_LogArchive_

Template for a LogArchive Step

Mostly borrowed from StageOut since they share a similar function

"""




from WMCore.WMSpec.Steps.Template import Template
from WMCore.WMSpec.Steps.Template import CoreHelper
from WMCore.WMSpec.ConfigSectionTree import nodeName



class LogArchiveStepHelper(CoreHelper):
    """
    _StageOutStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for StageOut tasks

    This is very similar to StageOut since they have essentially the same function

    """
    def addLog(self, infile):
        """
            Enqueues(sp?) a file to the StageOut step
            infile must be a PFN
            outfile must be a LFN
        """
        target = self.data.logs.section_("log%i" % self.data.filecount)
        target.input  = infile
        self.data.filecount += 1

    def disableRetries(self):
        """
            handy for testing, without the 10 minute retry loop
        """
        self.data.retryCount = 1
        self.data.retryDelay = 0

    def addOutputDestination(self, lfn):
        """
        Adds an out location to put a tarball of all logs
        """


class LogArchive(Template):
    """
    _LogArchive_

    Tools for creating a template LogArchive Step

    """

    def install(self, step):
        stepname = nodeName(step)
        step.stepType = "LogArchive"
        step.section_("logs")
        step.logcount = 0
        step.retryCount = 3
        step.retryDelay = 300

        # Create output to put logs in
        step.section_("output")
        step.output.section_("modules")
        step.output.modules.section_("logArchive")


    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return LogArchiveStepHelper(step)



#!/usr/bin/env python
"""
WorkQueueBackend

Interface to WorkQueue persistent storage
"""

from builtins import object
from math import ceil

from future.utils import viewitems

import json
import random
import time

from WMCore.Database.CMSCouch import CouchServer, CouchNotFoundError, Document
from WMCore.Lexicon import sanitizeURL
from WMCore.WMSpec.WMWorkload import WMWorkloadHelper
from WMCore.WorkQueue.DataStructs.CouchWorkQueueElement import CouchWorkQueueElement, fixElementConflicts
from WMCore.WorkQueue.DataStructs.WorkQueueElement import possibleSites
from WMCore.WorkQueue.WorkQueueExceptions import WorkQueueNoMatchingElements, WorkQueueError


def formatReply(answer, *items):
    """Take reply from couch bulk api and format labeling errors etc
    """
    result, errors = [], []
    for ans in answer:
        if 'error' in ans:
            errors.append(ans)
            continue
        for item in items:
            if item.id == ans['id']:
                item.rev = ans['rev']
                result.append(item)
                break
    return result, errors


def sortAvailableElements(elementsList):
    """
    Given a list of workqueue elements dictionary, this function will
    sort them in place, first by their creation time; secondly by their
    priority. In other words, higher priority and older requests will
    be first in the list
    :param elementsList: a list of elements dictionary
    :return: nothing, list is updated in place
    """
    elementsList.sort(key=lambda element: element['CreationTime'])
    elementsList.sort(key=lambda element: element['Priority'], reverse=True)


class WorkQueueBackend(object):
    """
    Represents persistent storage for WorkQueue
    """

    def __init__(self, db_url, db_name='workqueue',
                 inbox_name=None, parentQueue=None,
                 queueUrl=None, logger=None):
        if logger:
            self.logger = logger
        else:
            import logging
            self.logger = logging

        if inbox_name is None:
            inbox_name = "%s_inbox" % db_name

        self.server = CouchServer(db_url)
        self.parentCouchUrlWithAuth = parentQueue
        if parentQueue:
            self.parentCouchUrl = sanitizeURL(parentQueue)['url']
        else:
            self.parentCouchUrl = None
        self.db = self.server.connectDatabase(db_name, create=False, size=10000)
        self.hostWithAuth = db_url
        self.inbox = self.server.connectDatabase(inbox_name, create=False, size=10000)
        self.queueUrlWithAuth = queueUrl or (db_url + '/' + db_name)
        self.queueUrl = sanitizeURL(queueUrl or (db_url + '/' + db_name))['url']
        self.eleKey = 'WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'

    def forceQueueSync(self):
        """Setup CouchDB replications - used only in tests"""
        self.pullFromParent(continuous=True)
        self.sendToParent(continuous=True)

    def pullFromParent(self, continuous=True, cancel=False):
        """Replicate from parent couch - blocking: used only in unit tests"""
        try:
            if self.parentCouchUrlWithAuth and self.queueUrlWithAuth:
                self.logger.info("Forcing pullFromParent from parentCouch: %s to queueUrl %s/%s",
                                 self.parentCouchUrlWithAuth, self.queueUrlWithAuth, self.inbox.name)
                self.server.replicate(source=self.parentCouchUrlWithAuth,
                                      destination="%s/%s" % (self.hostWithAuth, self.inbox.name),
                                      filter='WorkQueue/queueFilter',
                                      query_params={'childUrl': self.queueUrl,
                                                    'parentUrl': self.parentCouchUrl},
                                      continuous=continuous,
                                      cancel=cancel,
                                      sleepSecs=6)
        except Exception as ex:
            self.logger.warning('Replication from %s failed: %s' % (self.parentCouchUrlWithAuth, str(ex)))

    def sendToParent(self, continuous=True, cancel=False):
        """Replicate to parent couch - blocking: used only int test"""
        try:
            if self.parentCouchUrlWithAuth and self.queueUrlWithAuth:
                self.logger.info("Forcing sendToParent from queueUrl %s/%s to parentCouch: %s",
                                 self.queueUrlWithAuth, self.inbox.name, self.parentCouchUrlWithAuth)
                self.server.replicate(source="%s" % self.inbox.name,
                                      destination=self.parentCouchUrlWithAuth,
                                      filter='WorkQueue/queueFilter',
                                      query_params={'childUrl': self.queueUrl,
                                                    'parentUrl': self.parentCouchUrl},
                                      continuous=continuous,
                                      cancel=cancel)
        except Exception as ex:
            self.logger.warning('Replication to %s failed: %s' % (self.parentCouchUrlWithAuth, str(ex)))

    def getElementsForSplitting(self):
        """Returns the elements from the inbox that need to be split,
        if WorkflowName specified only return elements to split for that workflow"""
        elements = self.getInboxElements(status='Negotiating')
        specs = {}  # cache as may have multiple elements for same spec
        for ele in elements:
            if ele['RequestName'] not in specs:
                wmspec = WMWorkloadHelper()
                wmspec.load(self.parentCouchUrlWithAuth + "/%s/spec" % ele['RequestName'])
                specs[ele['RequestName']] = wmspec
            ele['WMSpec'] = specs[ele['RequestName']]
        del specs
        return elements

    def insertWMSpec(self, wmspec):
        """
        Insert WMSpec to backend
        """
        # Can't save spec to inbox, it needs to be visible to child queues
        # Can't save empty dict so add dummy variable
        dummyValues = {'name': wmspec.name()}
        # change specUrl in spec before saving (otherwise it points to previous url)
        wmspec.setSpecUrl(self.db['host'] + "/%s/%s/spec" % (self.db.name, wmspec.name()))
        return wmspec.saveCouch(self.hostWithAuth, self.db.name, dummyValues)

    def getWMSpec(self, name):
        """Get the spec"""
        wmspec = WMWorkloadHelper()
        wmspec.load(self.hostWithAuth + "/%s/%s/spec" % (self.db.name, name))
        return wmspec

    def insertElements(self, units, parent=None):
        """
        Insert element to database

        @param parent is the parent WorkQueueObject these element's belong to.
                                            i.e. a workflow which has been split
        """
        if not units:
            return []
        # store spec file separately - assume all elements share same spec
        self.insertWMSpec(units[0]['WMSpec'])
        newUnitsInserted = []
        for unit in units:
            # cast to couch
            if not isinstance(unit, CouchWorkQueueElement):
                unit = CouchWorkQueueElement(self.db, elementParams=dict(unit))

            if parent:
                unit['ParentQueueId'] = parent.id
                unit['TeamName'] = parent['TeamName']
                unit['WMBSUrl'] = parent['WMBSUrl']

            if unit._couch.documentExists(unit.id):
                self.logger.info('Element "%s" already exists, skip insertion.' % unit.id)
                continue

            newUnitsInserted.append(unit)
            unit.save()
            # FIXME: this is not performing bulk request, but single document commits(!)
            unit._couch.commit()

        return newUnitsInserted

    def createWork(self, spec, **kwargs):
        """Return the Inbox element for this spec.

        This does not persist it to the database.
        """
        kwargs.update({'WMSpec': spec,
                       'RequestName': spec.name(),
                       'StartPolicy': spec.startPolicyParameters(),
                       'EndPolicy': spec.endPolicyParameters(),
                       'OpenForNewData': True
                       })
        unit = CouchWorkQueueElement(self.inbox, elementParams=kwargs)
        unit.id = spec.name()
        return unit

    def getElements(self, status=None, elementIDs=None, returnIdOnly=False,
                    db=None, loadSpec=False, WorkflowName=None, **elementFilters):
        """Return elements that match requirements

        status, elementIDs & filters are 'AND'ed together to filter elements.
        returnIdOnly causes the element not to be loaded and only the id returned
        db is used to specify which database to return from
        loadSpec causes the workflow for each spec to be loaded.
        WorkflowName may be used in the place of RequestName
        """
        key = []
        if not db:
            db = self.db
        if elementFilters.get('RequestName') and not WorkflowName:
            WorkflowName = elementFilters.pop('RequestName')

        if elementIDs:
            if elementFilters or status or returnIdOnly:
                msg = "Can't specify extra filters (or return id's) when using element id's with getElements()"
                raise ValueError(msg)
            elements = [CouchWorkQueueElement(db, i).load() for i in elementIDs]
        else:
            options = {'include_docs': True, 'filter': elementFilters, 'idOnly': returnIdOnly, 'reduce': False}
            # filter on workflow or status if possible
            filterName = 'elementsByWorkflow'
            if WorkflowName:
                key.append(WorkflowName)
            elif status:
                filterName = 'elementsByStatus'
                key.append(status)
            elif elementFilters.get('SubscriptionId'):
                key.append(elementFilters['SubscriptionId'])
                filterName = 'elementsBySubscription'
            # add given params to filters
            if status:
                options['filter']['Status'] = status
            if WorkflowName:
                options['filter']['RequestName'] = WorkflowName

            view = db.loadList('WorkQueue', 'filter', filterName, options, key)
            view = json.loads(view)
            if returnIdOnly:
                return view
            elements = [CouchWorkQueueElement.fromDocument(db, row) for row in view]

        if loadSpec:
            specs = {}  # cache as may have multiple elements for same spec
            for ele in elements:
                if ele['RequestName'] not in specs:
                    wmspec = self.getWMSpec(ele['RequestName'])
                    specs[ele['RequestName']] = wmspec
                ele['WMSpec'] = specs[ele['RequestName']]
            del specs
        return elements

    def getInboxElements(self, *args, **kwargs):
        """
        Return elements from Inbox, supports same semantics as getElements()
        """
        return self.getElements(*args, db=self.inbox, **kwargs)

    def getElementsForWorkflow(self, workflow):
        """Get elements for a workflow"""
        elements = self.db.loadView('WorkQueue', 'elementsByWorkflow',
                                    {'key': workflow, 'include_docs': True, 'reduce': False})
        return [CouchWorkQueueElement.fromDocument(self.db,
                                                   x['doc'])
                for x in elements.get('rows', [])]

    def getElementsForParent(self, parent):
        """Get elements with the given parent"""
        elements = self.db.loadView('WorkQueue', 'elementsByParent', {'key': parent.id, 'include_docs': True})
        return [CouchWorkQueueElement.fromDocument(self.db,
                                                   x['doc'])
                for x in elements.get('rows', [])]

    def saveElements(self, *elements):
        """Persist elements

        Returns elements successfully saved, user must verify to catch errors
        """
        result = []
        if not elements:
            return result
        for element in elements:
            element.save()
        answer = elements[0]._couch.commit()
        result, failures = formatReply(answer, *elements)
        msg = 'Couch error saving element: "%s", error "%s", reason "%s"'
        for failed in failures:
            self.logger.error(msg % (failed['id'], failed['error'], failed['reason']))
        return result

    def _raiseConflictErrorAndLog(self, conflictIDs, updatedParams, dbName="workqueue"):
        errorMsg = "Need to update this element manually from %s\n ids:%s\n, parameters:%s\n" % (
            dbName, conflictIDs, updatedParams)
        self.logger.error(errorMsg)
        raise WorkQueueError(errorMsg)

    def updateElements(self, *elementIds, **updatedParams):
        """Update given element's (identified by id) with new parameters"""
        if not elementIds:
            return
        eleParams = {}
        eleParams[self.eleKey] = updatedParams
        conflictIDs = self.db.updateBulkDocumentsWithConflictHandle(elementIds, eleParams)
        if conflictIDs:
            self._raiseConflictErrorAndLog(conflictIDs, updatedParams)
        return

    def updateInboxElements(self, *elementIds, **updatedParams):
        """Update given inbox element's (identified by id) with new parameters"""
        if not elementIds:
            return
        eleParams = {}
        eleParams[self.eleKey] = updatedParams
        conflictIDs = self.inbox.updateBulkDocumentsWithConflictHandle(elementIds, eleParams)
        if conflictIDs:
            self._raiseConflictErrorAndLog(conflictIDs, updatedParams, "workqueue_inbox")
        return

    def deleteElements(self, *elements):
        """Delete elements"""
        if not elements:
            return
        specs = {}
        for i in elements:
            i.delete()
            specs[i['RequestName']] = None
        answer = elements[0]._couch.commit()
        _, failures = formatReply(answer, *elements)
        msg = 'Couch error deleting element: "%s", error "%s", reason "%s"'
        for failed in failures:
            # only count delete as failed if document still exists
            if elements[0]._couch.documentExists(failed['id']):
                self.logger.error(msg % (failed['id'], failed['error'], failed['reason']))
        # delete specs if no longer used
        for wf in specs:
            try:
                if not self.db.loadView('WorkQueue', 'elementsByWorkflow',
                                        {'key': wf, 'limit': 1, 'reduce': False})['rows']:
                    self.db.delete_doc(wf)
            except CouchNotFoundError:
                pass

    def calculateAvailableWork(self, thresholds, siteJobCounts):
        """
        A short version of the `availableWork` method, which is used only to calculate
        the amount of work already available at the local workqueue.
        :param thresholds: a dictionary key'ed by the site name, values representing the
            maximum number of jobs allowed at that site.
        :param siteJobCounts: a dictionary-of-dictionaries key'ed by the site name; value
            is a dictionary with the number of jobs running at a given priority.
        :return: a tuple with the elements accepted and an overview of job counts per site
        """
        # NOTE: this method can be less verbose as well
        elements = []
        # If there are no sites, punt early.
        if not thresholds:
            self.logger.error("No thresholds is set: Please check")
            return elements, siteJobCounts

        self.logger.info("Calculating available work from queue %s", self.queueUrl)

        options = {}
        options['include_docs'] = True
        options['descending'] = True
        options['resources'] = thresholds
        options['num_elem'] = 9999999  # magic number!
        result = self.db.loadList('WorkQueue', 'workRestrictions', 'availableByPriority', options)
        result = json.loads(result)
        self.logger.info("Retrieved %d elements from workRestrictions list for: %s",
                         len(result), self.queueUrl)

        # Convert python dictionary into Couch WQE objects
        # And sort them by creation time and priority, such that highest priority and
        # oldest elements come first in the list
        sortedElements = []
        for item in result:
            element = CouchWorkQueueElement.fromDocument(self.db, item)
            sortedElements.append(element)
        sortAvailableElements(sortedElements)

        for element in sortedElements:
            commonSites = possibleSites(element)
            prio = element['Priority']
            # shuffle list of common sites all the time to give everyone the same chance
            random.shuffle(commonSites)
            possibleSite = None
            for site in commonSites:
                if site in thresholds:
                    # Count the number of jobs currently running of greater priority, if they
                    # are less than the site thresholds, then accept this element
                    curJobCount = sum([x[1] if x[0] >= prio else 0 for x in viewitems(siteJobCounts.get(site, {}))])
                    self.logger.debug("Job Count: %s, site: %s thresholds: %s", curJobCount, site, thresholds[site])
                    if curJobCount < thresholds[site]:
                        possibleSite = site
                        break

            if possibleSite:
                self.logger.debug("Meant to accept workflow: %s, with prio: %s, element id: %s, for site: %s",
                                  element['RequestName'], prio, element.id, possibleSite)
                elements.append(element)
                siteJobCounts.setdefault(possibleSite, {})
                siteJobCounts[possibleSite][prio] = siteJobCounts[possibleSite].setdefault(prio, 0) + \
                                                    element['Jobs'] * element.get('blowupFactor', 1.0)
            else:
                self.logger.debug("No available resources for %s with localdoc id %s",
                                  element['RequestName'], element.id)

        self.logger.info("And %d elements passed location and siteJobCounts restrictions for: %s",
                         len(elements), self.queueUrl)
        return elements, siteJobCounts

    def availableWork(self, thresholds, siteJobCounts, team=None, excludeWorkflows=None,
                      numElems=1000, rowsPerSlice=1000, maxRows=1000):
        """
        Get work - either from local or global queue - which is available to be run.

        :param thresholds: a dictionary key'ed by the site name, values representing the
            maximum number of jobs allowed at that site.
        :param siteJobCounts: a dictionary-of-dictionaries key'ed by the site name; value
            is a dictionary with the number of jobs running at a given priority.
        :param team: a string with the team name we want to pull work for
        :param excludeWorkflows: list of (aborted) workflows that should not be accepted
        :param numElems: integer with the maximum number of elements to be accepted.
        :param rowsPerSlice: integer defining the amount of rows for each slice (slices
             of a couchdb view request).
        :param maxRows: maximum number of available elements (rows) to be considered
            when pulling work down to the agent.
        :return: a tuple with the elements accepted and an overview of job counts per site
        """
        excludeWorkflows = excludeWorkflows or []
        acceptedElems = []
        # If there are no sites, punt early.
        if not thresholds:
            self.logger.error("No thresholds is set: Please check")
            return acceptedElems, siteJobCounts

        self.logger.info("Current siteJobCounts:")
        for site, jobsByPrio in viewitems(siteJobCounts):
            self.logger.info("    %s : %s", site, jobsByPrio)

        # Find out how many elements are in Available status
        numAvail = self.queueLength()
        self.logger.info("Current amount of WQEs in Available status: %s", numAvail)

        self.logger.info("Getting up to %d available work from %s", numElems, self.queueUrl)
        self.logger.info("  for team name: %s", team)
        self.logger.info("  with excludeWorkflows count: %s", len(excludeWorkflows))
        self.logger.debug("  with excludeWorkflows: %s", excludeWorkflows)
        self.logger.info("  for thresholds: %s", thresholds)

        options = {}
        options['include_docs'] = True
        options['descending'] = True
        options['resources'] = thresholds
        options['limit'] = rowsPerSlice
        # FIXME: num_elem option can likely be deprecated, but it needs synchronization
        # between agents and global workqueue... for now, make sure it can return the slice size
        options['num_elem'] = rowsPerSlice
        if team:
            options['team'] = team

        # Fetch workqueue elements in slices, using the CouchDB "limit" and "skip"
        # options for couch views. Conditions to stop this loop are:
        #  a) stop once total_rows is reached (exhausted all available elements)
        #  b) hit maximum allowed elements/rows to be considered for data acquisition (maxRows)
        #  c) or, once the targeted number of elements has been accepted (numElems)
        numSlices = ceil(numAvail / rowsPerSlice)
        numSlices = min(numSlices, int(maxRows / rowsPerSlice))
        for sliceNum in range(numSlices):
            # documents to skip as a function of the slice number
            options['skip'] = sliceNum * rowsPerSlice
            self.logger.info("  for slice: %s, with rows range [%s - %s]",
                             sliceNum, options['skip'], options['skip'] + options['limit'])

            result = self.db.loadList('WorkQueue', 'workRestrictions', 'availableByPriority', options)
            # now check the remaining restrictions and priority
            wqeSlots = numElems - len(acceptedElems)
            elems = self._evalAvailableWork(json.loads(result), thresholds, siteJobCounts,
                                            excludeWorkflows, wqeSlots)
            acceptedElems.extend(elems)
            if len(acceptedElems) >= numElems:
                msg = f"Reached maximum number of elements to be accepted, "
                msg += f"configured to: {numElems}, from queue: {self.queueUrl}"
                self.logger.info(msg)
                break

        self.logger.info("Total of %d elements passed location and siteJobCounts restrictions for: %s",
                         len(acceptedElems), self.queueUrl)
        return acceptedElems, siteJobCounts

    def _evalAvailableWork(self, listElems, thresholds, siteJobCounts,
                           excludeWorkflows, numElems):
        """
        Evaluate work available in workqueue and decide whether it can be
        accepted or not.

        :param listElems: list of dictionaries that correspond to the workqueue elements.
        :param thresholds: a dictionary key'ed by the site name, values representing the
            maximum number of jobs allowed at that site.
        :param siteJobCounts: a dictionary-of-dictionaries key'ed by the site name; value
            is a dictionary with the number of jobs running at a given priority.
            NOTE that it is updated in place.
        :param excludeWorkflows: list of (aborted) workflows that should not be accepted
        :param numElems: integer with the maximum number of elements to be accepted (default
            to a very large number when pulling work from local queue, read unlimited)
        :return: a tuple with the elements accepted and an overview of job counts per site
        """
        elems = []
        self.logger.info("Retrieved %d elements from workRestrictions list for: %s",
                         len(listElems), self.queueUrl)
        # Convert python dictionary into Couch WQE objects, skipping aborted workflows
        # And sort them by creation time and priority, such that highest priority and
        # oldest elements come first in the list
        sortedElements = []
        for i in listElems:
            element = CouchWorkQueueElement.fromDocument(self.db, i)
            # make sure not to acquire work for aborted or force-completed workflows
            if element['RequestName'] in excludeWorkflows:
                msg = "Skipping aborted/force-completed workflow: %s, work id: %s"
                self.logger.info(msg, element['RequestName'], element._id)
            elif element['Status'] != 'Available':
                # Extra safety mechanism, see https://github.com/dmwm/WMCore/pull/12050
                msg = "Skipping element in unwanted status: %s, work id: %s"
                self.logger.warning(msg, element['Status'], element._id)
            else:
                sortedElements.append(element)
        sortAvailableElements(sortedElements)

        for element in sortedElements:
            if numElems <= 0:
                # it means we accepted the configured number of elements
                break
            commonSites = possibleSites(element)
            prio = element['Priority']
            # shuffle list of common sites all the time to give everyone the same chance
            random.shuffle(commonSites)
            possibleSite = None
            for site in commonSites:
                if site in thresholds:
                    # Count the number of jobs currently running of greater priority, if they
                    # are less than the site thresholds, then accept this element
                    curJobCount = sum([x[1] if x[0] >= prio else 0 for x in viewitems(siteJobCounts.get(site, {}))])
                    self.logger.debug("Job Count: %s, site: %s thresholds: %s",
                                      curJobCount, site, thresholds[site])
                    if curJobCount < thresholds[site]:
                        possibleSite = site
                        break

            if possibleSite:
                self.logger.info("Accepting workflow: %s, with prio: %s, element id: %s, for site: %s",
                                 element['RequestName'], prio, element.id, possibleSite)
                numElems -= 1
                elems.append(element)
                siteJobCounts.setdefault(possibleSite, {})
                siteJobCounts[possibleSite][prio] = siteJobCounts[possibleSite].setdefault(prio, 0) + \
                                                    element['Jobs'] * element.get('blowupFactor', 1.0)
            else:
                self.logger.debug("No available resources for %s with doc id %s",
                                  element['RequestName'], element.id)

        self.logger.info("And %d elements passed location and siteJobCounts restrictions for: %s",
                         len(elems), self.queueUrl)
        return elems

    def getActiveData(self):
        """Get data items we have work in the queue for"""
        data = self.db.loadView('WorkQueue', 'activeData', {'reduce': True, 'group': True})
        return [{'dbs_url': x['key'][0],
                 'name': x['key'][1]} for x in data.get('rows', [])]

    def getActiveParentData(self):
        """Get data items we have work in the queue for with parent"""
        data = self.db.loadView('WorkQueue', 'activeParentData', {'reduce': True, 'group': True})
        return [{'dbs_url': x['key'][0],
                 'name': x['key'][1]} for x in data.get('rows', [])]

    def getActivePileupData(self):
        """Get data items we have work in the queue for with pileup"""
        data = self.db.loadView('WorkQueue', 'activePileupData', {'reduce': True, 'group': True})
        return [{'dbs_url': x['key'][0],
                 'name': x['key'][1]} for x in data.get('rows', [])]

    def getElementsForData(self, data):
        """Get active elements for this dbs & data combo"""
        elements = self.db.loadView('WorkQueue', 'elementsByData', {'key': data, 'include_docs': True})
        return [CouchWorkQueueElement.fromDocument(self.db,
                                                   x['doc'])
                for x in elements.get('rows', [])]

    def getElementsForParentData(self, data):
        """Get active elements for this data """
        elements = self.db.loadView('WorkQueue', 'elementsByParentData', {'key': data, 'include_docs': True})
        return [CouchWorkQueueElement.fromDocument(self.db,
                                                   x['doc'])
                for x in elements.get('rows', [])]

    def getElementsForPileupData(self, data):
        """Get active elements for this data """
        elements = self.db.loadView('WorkQueue', 'elementsByPileupData', {'key': data, 'include_docs': True})
        return [CouchWorkQueueElement.fromDocument(self.db,
                                                   x['doc'])
                for x in elements.get('rows', [])]

    def isAvailable(self):
        """Is the server available, i.e. up and not compacting"""
        try:
            compacting = self.db.info()['compact_running']
            if compacting:
                self.logger.info("CouchDB compacting - try again later.")
                return False
        except Exception as ex:
            self.logger.error("CouchDB unavailable: %s" % str(ex))
            return False
        return True

    def getWorkflows(self, includeInbox=False, includeSpecs=False):
        """Returns workflows known to workqueue"""
        result = set([x['key'] for x in self.db.loadView('WorkQueue', 'elementsByWorkflow', {'group': True})['rows']])
        if includeInbox:
            result = result | set(
                    [x['key'] for x in self.inbox.loadView('WorkQueue', 'elementsByWorkflow', {'group': True})['rows']])
        if includeSpecs:
            result = result | set([x['key'] for x in self.db.loadView('WorkQueue', 'specsByWorkflow')['rows']])
        return list(result)

    def queueLength(self):
        """Return number of available elements"""
        return self.db.loadView('WorkQueue', 'availableByPriority', {'limit': 0})['total_rows']

    def fixConflicts(self):
        """Fix elements in conflict

        Each local queue runs this to resolve its conflicts with global,
        resolution propagates up to global.

        Conflicting elements are merged into one element with others deleted.

        This will fail if elements are modified during the resolution -
        if this happens rerun.
        """
        for db in [self.inbox, self.db]:
            for row in db.loadView('WorkQueue', 'conflicts')['rows']:
                elementId = row['id']
                try:
                    conflicting_elements = [CouchWorkQueueElement.fromDocument(db, db.document(elementId, rev)) \
                                            for rev in row['value']]
                    fixed_elements = fixElementConflicts(*conflicting_elements)
                    if self.saveElements(fixed_elements[0]):
                        self.saveElements(*fixed_elements[1:])  # delete others (if merged value update accepted)
                except Exception as ex:
                    self.logger.error("Error resolving conflict for %s: %s" % (elementId, str(ex)))

    def recordTaskActivity(self, taskname, comment=''):
        """Record a task for monitoring"""
        try:
            record = self.db.document('task_activity')
        except CouchNotFoundError:
            record = Document('task_activity')
        record.setdefault('tasks', {})
        record['tasks'].setdefault(taskname, {})
        record['tasks'][taskname]['timestamp'] = time.time()
        record['tasks'][taskname]['comment'] = comment
        try:
            self.db.commitOne(record)
        except Exception as ex:
            self.logger.error("Unable to update task %s freshness: %s" % (taskname, str(ex)))

    def getWMBSInjectStatus(self, request=None):
        """
        This service only provided by global queue except on draining agent
        """
        options = {'group': True, 'reduce': True}
        if request:
            options.update(key=request)
        data = self.db.loadView('WorkQueue', 'wmbsInjectStatusByRequest', options)
        if request:
            if data['rows']:
                injectionStatus = data['rows'][0]['value']
                inboxElement = self.getInboxElements(WorkflowName=request)
                requestOpen = inboxElement[0].get('OpenForNewData', False) if inboxElement else False
                return injectionStatus and not requestOpen
            else:
                raise WorkQueueNoMatchingElements("%s not found" % request)
        else:
            injectionStatus = dict((x['key'], x['value']) for x in data.get('rows', []))
            finalInjectionStatus = []
            for request in injectionStatus:
                inboxElement = self.getInboxElements(WorkflowName=request)
                requestOpen = inboxElement[0].get('OpenForNewData', False) if inboxElement else False
                finalInjectionStatus.append({request: injectionStatus[request] and not requestOpen})

            return finalInjectionStatus

    def getWorkflowNames(self, inboxFlag=False):
        """Get workflow names from workqueue db"""
        if inboxFlag:
            db = self.inbox
        else:
            db = self.db
        data = db.loadView('WorkQueue', 'elementsByWorkflow',
                           {'stale': "update_after", 'reduce': True, 'group': True})
        return [x['key'] for x in data.get('rows', [])]

    def deleteWQElementsByWorkflow(self, workflowNames):
        """
        delete workqueue elements belongs to given workflow names
        it doen't check the status of workflow so need to be careful to use this.
        Pass only workflows which has the end status
        """
        deleted = 0
        dbs = [self.db, self.inbox]
        if not isinstance(workflowNames, list):
            workflowNames = [workflowNames]

        if len(workflowNames) == 0:
            return deleted
        options = {}
        options["stale"] = "update_after"
        options["reduce"] = False

        idsByWflow = {}
        for couchdb in dbs:
            result = couchdb.loadView("WorkQueue", "elementsByWorkflow", options, workflowNames)
            for entry in result["rows"]:
                idsByWflow.setdefault(entry['key'], [])
                idsByWflow[entry['key']].append(entry['id'])
            for wflow, docIds in viewitems(idsByWflow):
                self.logger.info("Going to delete %d documents in *%s* db for workflow: %s. Doc IDs: %s",
                                 len(docIds), couchdb.name, wflow, docIds)
                try:
                    couchdb.bulkDeleteByIDs(docIds)
                except CouchNotFoundError as exc:
                    self.logger.error("Failed to find one of the documents. Error: %s", str(exc))
                deleted += len(docIds)
        # delete the workflow with spec from workqueue db
        for wf in workflowNames:
            self.db.delete_doc(wf)
        return deleted



from __future__ import print_function

# HOWTO run:
# 1) set up WMCore
# 2) db init
#    bin/wmcore-db-init --config src/python/WMCore/examples/WorkQueue/GlobalWQConfig.py --create --modules=WMCore.WorkQueue.Database,WMCore.MsgService,WMCore.Agent.Database
# 3) run compoments:
#    bin/wmcoreD --start --config src/python/WMCore/examples/WorkQueue/MonitorServiceConfig.py
# 4) python __file__
#     -> this prints out html file, normally resulting html snippet will be used ...


from builtins import range
from future import standard_library
standard_library.install_aliases()

import json
import random
import urllib.parse

from WMCore.WorkQueue.DataStructs.WorkQueueElement import STATES


# random colour chooser (or use matplotlib colourmap)
def random_colour():
    return '#%02x%02x%02x'%(random.randint(0,255),random.randint(0,255),random.randint(0,255))


# output file name
FILE_NAME = "PlotFairyBarChart.html"
PLOT_SERVICE_URL = "http://localhost:8888"
DATA_SOURCE_URL = "http://localhost:8888/workqueue/elementsinfo"


pageTemplate = \
"""
<html>\n<body>
<p>
<img src=%(url)s/plotfairy/plot/?type=%(plotType)s&data=%(hardCodedData)s>
</p>
<br><br>
</body>\n</html>
"""


# hard-coded data
values = [random.randint(10, 500) for i in range(7)] # statuses at sites ...
series = [{"label": x, "colour":  random_colour(), "values": values} for x in STATES]
plotDefinition = \
{
    "title": "Job elements status",
    "series": series,
    "shadow": True,
    "sort": "value",
    # don't know how to have only text labels on x-axis (e.g. sites names) - format
    "xaxis": {"label": "status at sites", "min": 10, "format": "si", "width": 20, "bins": 5},
    "yaxis": {"label": "thousands of jobs", "format": "num"},
    "legend": "right"
}
plotDefinition["title"] = "Element status - hard-coded"
hardCodedPlotData = urllib.parse.quote(json.dumps(plotDefinition, ensure_ascii = True))



# TODO
# get data from the system (URL) ... should have particular visualisation request ...
# see PlotFairyPieChart for reference ...
# retrieve data from the system
# URL http://localhost:8888/workqueue/elementsinfo

# write result file / generate result html snippet
outputFile = open(FILE_NAME, 'w')
finalPage = pageTemplate % {"url": PLOT_SERVICE_URL, "plotType": "Bar",
                            "hardCodedData": hardCodedPlotData}
outputFile.write(finalPage)
outputFile.close()
print("result html page '%s' written" % FILE_NAME)



"""
Timers_t module provide unit tests for Timer module
Unittests for Utilities functions
"""

# system modules
import unittest

# WMCore modules
from Utils.Timers import gmtimeSeconds, encodeTimestamp, decodeTimestamp


class TimersTests(unittest.TestCase):
    """
    unittest for Timers functions
    """
    def testTimingFunctions(self):
        "Test timing functions"
        gmtime = gmtimeSeconds()
        self.assertEqual(gmtime > 0, True)
        self.assertEqual(len("{}".format(gmtime)), 10)

        tst = 1674477244
        expect = '2023-01-23T12:34:04Z'
        val = encodeTimestamp(tst)
        self.assertEqual(val, expect)
        val = decodeTimestamp(val)
        self.assertEqual(val, tst)


if __name__ == '__main__':
    unittest.main()



#decorator import for RESTServer setup
from WMQuality.TestInitCouchApp import TestInitCouchApp
from WMQuality.WebTools.RESTBaseUnitTest import RESTBaseUnitTest

class RESTAndCouchUnitTest(RESTBaseUnitTest):

    def setUp(self):
        # default set
        if self.schemaModules:
            self.testInit = TestInitCouchApp(__file__)
        RESTBaseUnitTest.__init__(self)



#!/usr/bin/env python
"""
_JobWorkUnit_t_

Unit tests for the interaction of WMBS Job and WorkUnit
"""
from __future__ import absolute_import, division, print_function

import unittest

from WMCore.DataStructs.Run import Run
from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset as Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.WorkUnit import WorkUnit
from WMCore.WMBS.Workflow import Workflow
from WMCore_t.WMBS_t.JobTestBase import JobTestBase


class JobWorkUnitTest(JobTestBase):
    """
    Create jobs and test that associated WorkUnits are correct
    """

    def notestJobCreatesWorkUnits(self):
        """
        Create a single jobs and test that both work units are created
        """

        self.createSingleJobWorkflow()

        testRunLumi = Run(1, 45)
        loadByFRL = WorkUnit(taskID=self.testWorkflow.id, fileid=self.testFileA['id'], runLumi=testRunLumi)
        loadByFRL.load()

        self.assertEqual(loadByFRL['last_unit_count'], 2)  # Two lumis in this job
        self.assertGreater(loadByFRL['id'], 0)

        testRunLumi = Run(1, 46)
        loadByFRL = WorkUnit(taskID=self.testWorkflow.id, fileid=self.testFileB['id'], runLumi=testRunLumi)
        loadByFRL.load()

        self.assertEqual(loadByFRL['last_unit_count'], 2)  # Two lumis in this job
        self.assertGreater(loadByFRL['id'], 0)

        return

    def notestCreateDeleteExists(self):
        """
        Create and then delete a job and workflow.  Use the workunit class's exists() method to
        determine if the workunit has been written to the database before the job is
        created, after the job has been created, and after the workflow has been deleted.
        """

        testWorkflow = Workflow(spec="spec.xml", owner="Simon", name="wf001", task="Test")
        testWorkflow.create()

        testWMBSFileset = Fileset(name="TestFileset")
        testWMBSFileset.create()

        testSubscription = Subscription(fileset=testWMBSFileset, workflow=testWorkflow)
        testSubscription.create()

        testJobGroup = JobGroup(subscription=testSubscription)
        testJobGroup.create()

        testFileA = File(lfn="/this/is/a/lfnA", size=1024, events=10)
        testFileA.addRun(Run(1, *[45]))
        testFileB = File(lfn="/this/is/a/lfnB", size=1024, events=10)
        testFileB.addRun(Run(1, *[46]))

        testFileA.create()
        testFileB.create()

        testJob = Job(name="TestJob", files=[testFileA, testFileB])
        testWU1 = WorkUnit(taskID=testWorkflow.id, fileid=testFileA['id'], runLumi=Run(1, *[45]))
        testWU2 = WorkUnit(taskID=testWorkflow.id, fileid=testFileB['id'], runLumi=Run(1, *[46]))

        self.assertFalse(testWU1.exists(), "WorkUnit exists before job was created")
        self.assertFalse(testWU2.exists(), "WorkUnit exists before job was created")

        testJob.create(group=testJobGroup)

        self.assertTrue(testWU1.exists(), "WorkUnit does not exist after job was created")
        self.assertTrue(testWU2.exists(), "WorkUnit does not exist after job was created")

        testJob.delete()

        self.assertTrue(testWU1.exists(), "WorkUnit does not exist after job is deleted")
        self.assertTrue(testWU2.exists(), "WorkUnit does not exist after job is deleted")

        testWorkflow.delete()

        self.assertFalse(testWU1.exists(), "WorkUnit exists after workflow is deleted")
        self.assertFalse(testWU2.exists(), "WorkUnit exists after workflow is deleted")

        return


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_GetSemiFinishedTasks_

Oracle implementation of Subscriptions.GetSemiFinishedTasks

Created on Nov 7, 2012

@author: dballest
"""

from WMCore.WMBS.MySQL.Subscriptions.GetSemiFinishedTasks import GetSemiFinishedTasks as MySQLGetSemiFinishedTasks

class GetSemiFinishedTasks(MySQLGetSemiFinishedTasks):
    pass



#!/usr/bin/env python
"""
_HadoopImpl_

Implementation of StageOutImpl interface for Vanderbilt

"""

import os.path

from WMCore.Storage.Registry import registerStageOutImpl
from WMCore.Storage.StageOutImpl import StageOutImpl


class VandyImpl(StageOutImpl):
    '''
    _VandyImp_

    Implement the vandy interface
    '''

    # BASEDIR='/Users/brumgard/Documents/workspace/VandyStageOut/src/scripts'
    BASEDIR = '/usr/local/cms-stageout'

    def __init__(self, stagein=False):

        StageOutImpl.__init__(self, stagein)

        self._mkdirScript = os.path.join(VandyImpl.BASEDIR, 'vandyMkdir.sh')
        self._cpScript = os.path.join(VandyImpl.BASEDIR, 'vandyCp.sh')
        self._rmScript = os.path.join(VandyImpl.BASEDIR, 'vandyRm.sh')
        self._downloadScript = os.path.join(VandyImpl.BASEDIR, 'vandyDownload.sh')

    def createSourceName(self, protocol, pfn):

        """
        _createSourceName_

        uses pfn

        """

        return "%s" % pfn

    def createOutputDirectory(self, targetPFN):

        """
        _createOutputDirectory_

        Creates the directory for vanderbilt
        """

        command = "%s %s" % (self._mkdirScript, os.path.dirname(targetPFN))

        # Calls the parent execute command to invoke the script which should
        # throw a stage out error
        self.executeCommand(command)

    def createStageOutCommand(self, sourcePFN, targetPFN, options=None, checksums=None):

        """
        _createStageOutCommand_

        Build a shell command that will transfer the sourcePFN to the
        targetPFN.

        """
        if self.stageIn:
            return "%s %s %s" % (self._downloadScript, sourcePFN, targetPFN)
        else:
            return "%s %s %s" % (self._cpScript, sourcePFN, targetPFN)
        
    def createDebuggingCommand(self, sourcePFN, targetPFN, options=None, checksums=None):
        """
        Debug a failed vandy copy command for stageOut, without re-running it,
        providing information on the environment and the certifications

        :sourcePFN: str, PFN of the source file
        :targetPFN: str, destination PFN
        :options: str, additional options for copy command
        :checksums: dict, collect checksums according to the algorithms saved as keys
        """
        # Build the command for debugging purposes
        copyCommand = ""
        if self.stageIn:
            copyCommand += "%s %s %s" % (self._downloadScript, sourcePFN, targetPFN)
        else:
            copyCommand += "%s %s %s" % (self._cpScript, sourcePFN, targetPFN)

        result = self.debuggingTemplate.format(copy_command=copyCommand, source=sourcePFN, destination=targetPFN)
        return result

    def removeFile(self, pfnToRemove):

        """
        _removeFile_

        Removes the pfn.
        """

        command = "%s %s" % (self._rmScript, pfnToRemove)

        self.executeCommand(command)


# Registers the implementation
registerStageOutImpl("vandy", VandyImpl)



"""
Unit tests for the WMCore/MicroService/MSOutput/RelValPolicy.py module
"""
from __future__ import division, print_function

import json
import unittest

from WMCore.MicroService.MSOutput.RelValPolicy import RelValPolicy, RelValPolicyException


class RelValPolicyTests(unittest.TestCase):
    """Unit tests for the RelValPolicy module"""

    def setUp(self):
        """Basic setup for each unit test"""
        pass

    def tearDown(self):
        """Basic tear down operation when leaving each unit test"""
        pass

    def testBrokenTierPolicy(self):
        """Tests for the RelValPolicy class with a broken tier-based policy"""
        validDatatiers = ["GEN-SIM", "GEN", "SIM", "AOD"]
        validRSEs = ["rse_1", "rse_2", "rse_3"]
        timePol = [{"releaseType": "pre", "lifetimeSecs": 1},
                   {"releaseType": "default", "lifetimeSecs": 2}]

        # test output policies with a wrong data type
        for testPolicy in [None, {}, "blah", 123]:
            with self.assertRaises(RelValPolicyException):
                RelValPolicy(testPolicy, timePol, validDatatiers, validRSEs)

        # test internal structure with wrong data type for datatier
        testPolicy = [{"datatier": ["tier1", "tier2"], "destinations": ["rse_1", "rse_2"]},
                      {"datatier": "default", "destinations": ["rse_1"]}]
        with self.assertRaises(RelValPolicyException):
            RelValPolicy(testPolicy, timePol, validDatatiers, validRSEs)

        # test internal structure with wrong data type for destinations
        testPolicy = [{"datatier": "tier1", "destinations": "rse_1"},
                      {"datatier": "default", "destinations": ["rse_1"]}]
        with self.assertRaises(RelValPolicyException):
            RelValPolicy(testPolicy, timePol, validDatatiers, validRSEs)

        # test internal structure missing the required "default" key/value pair
        testPolicy = [{"datatier": "GEN", "destinations": ["rse_1", "rse_2"]}]
        with self.assertRaises(RelValPolicyException):
            RelValPolicy(testPolicy, timePol, validDatatiers, validRSEs)

    def testBrokenLifePolicy(self):
        """Tests for the RelValPolicy class with a broken lifetime-based policy"""
        validDatatiers = ["GEN-SIM", "GEN", "SIM", "AOD"]
        validRSEs = ["rse_1", "rse_2", "rse_3"]
        tierPolicy = [{"datatier": "GEN", "destinations": ["rse_1"]},
                      {"datatier": "default", "destinations": ["rse_2"]}]

        # test output lifetime policy with a wrong data type
        for testPolicy in [None, {}, "blah", 123]:
            with self.assertRaises(RelValPolicyException):
                RelValPolicy(tierPolicy, testPolicy, validDatatiers, validRSEs)

        # test internal structure with wrong data type for releaseType
        pol1 = [{"releaseType": 123, "lifetimeSecs": 1}]
        pol2 = [{"releaseType": None, "lifetimeSecs": 1}]
        pol3 = [{"releaseType": ["pre"], "lifetimeSecs": 1}, {"releaseType": "final", "lifetimeSecs": 2}]
        for lifePol in (pol1, pol2, pol3):
            with self.assertRaises(RelValPolicyException):
                RelValPolicy(tierPolicy, lifePol, validDatatiers, validRSEs)

        # test internal structure with wrong data type for lifetimeSecs
        pol1 = [{"releaseType": "pre", "lifetimeSecs": 1.5}]
        pol2 = [{"releaseType": "default", "lifetimeSecs": "1"}]
        pol3 = [{"releaseType": "pre", "lifetimeSecs": None}, {"releaseType": "final", "lifetimeSecs": 2}]
        pol4 = [{"releaseType": "pre", "lifetimeSecs": 0}, {"releaseType": "final", "lifetimeSecs": -10}]
        for lifePol in (pol1, pol2, pol3, pol4):
            with self.assertRaises(RelValPolicyException):
                RelValPolicy(tierPolicy, lifePol, validDatatiers, validRSEs)

        # test missing release type
        pol1 = [{"releaseType": "pre", "lifetimeSecs": 1}]
        pol2 = [{"releaseType": "default", "lifetimeSecs": 1}]
        pol3 = [{"releaseType": "pre", "lifetimeSecs": 1}, {"releaseType": "final", "lifetimeSecs": 2}]
        for lifePol in (pol1, pol2, pol3):
            with self.assertRaises(RelValPolicyException):
                RelValPolicy(tierPolicy, lifePol, validDatatiers, validRSEs)

    def testValidPolicy(self):
        """Tests for the RelValPolicy class with a valid policy"""
        validDatatiers = ["GEN-SIM", "GEN", "SIM", "AOD"]
        validRSEs = ["rse_1", "rse_2", "rse_3"]

        tierPolicy = [{"datatier": "SIM", "destinations": ["rse_1", "rse_2"]},
                      {"datatier": "GEN-SIM", "destinations": ["rse_1"]},
                      {"datatier": "default", "destinations": ["rse_2"]}]
        lifePolicy = [{"releaseType": "pre", "lifetimeSecs": 1},
                      {"releaseType": "default", "lifetimeSecs": 2}]

        policyObj = RelValPolicy(tierPolicy, lifePolicy, validDatatiers, validRSEs)

        # now test the method to get destinations for a given dataset (datatier)
        for policyItem in tierPolicy:
            resp = policyObj.getDestinationByDataset("/PD/ProcStr-v1/{}".format(policyItem['datatier']))
            self.assertEqual(resp, policyItem['destinations'])

        # and this should fallback to the 'default' case
        resp = policyObj.getDestinationByDataset("/PD/ProcStr-v1/BLAH")
        self.assertEqual(resp, ["rse_2"])

        # now test the release type lifetime policy using the default value
        for dsetName in {"/PD/CMSSW_1_2_3_patch1-ProcStr-v1/TIER",
                         "/PD/CMSSW_1_2_3_HLT1-ProcStr-v1/TIER",
                         "/PD/CMSSW_1_2_3-ProcStr-v1/TIER"}:
            self.assertEqual(policyObj.getLifetimeByDataset(dsetName), 2)
        # and a pre-release
        self.assertEqual(policyObj.getLifetimeByDataset("/PD/CMSSW_1_2_3_pre4-ProcStr-v1/TIER"), 1)

    def testStringification(self):
        """Test the stringification of the RelValPolicy object"""
        validDatatiers = []
        validRSEs = ["rse_2"]
        testPolicy = [{"datatier": "default", "destinations": ["rse_2"]}]
        lifePolicy = [{"releaseType": "pre", "lifetimeSecs": 3},
                      {"releaseType": "default", "lifetimeSecs": 12}]

        policyObj = str(RelValPolicy(testPolicy, lifePolicy, validDatatiers, validRSEs))

        self.assertTrue(isinstance(policyObj, str))
        policyObj = json.loads(policyObj)
        self.assertCountEqual(policyObj["originalTierPolicy"], testPolicy)
        self.assertCountEqual(policyObj["mappedTierPolicy"], {"default": ["rse_2"]})

        self.assertCountEqual(policyObj["originalLifetimePolicy"], lifePolicy)
        self.assertCountEqual(policyObj["mappedLifetimePolicy"], {"pre": 3, "default": 12})

    def testIsPreRelease(self):
        """Test the _isPreRelease method"""
        tierPolicy = [{"datatier": "SIM", "destinations": ["rse_1"]},
                      {"datatier": "default", "destinations": ["rse_1"]}]
        lifePolicy = [{"releaseType": "pre", "lifetimeSecs": 3},
                      {"releaseType": "default", "lifetimeSecs": 12}]

        policyObj = RelValPolicy(tierPolicy, lifePolicy, ["SIM"], ["rse_1"])

        # first test a broken dataset name
        for dsetName in ("", None), 123:
            with self.assertRaises(RuntimeError):
                policyObj._isPreRelease(dsetName)

        for dsetName in {"/PD/CMSSW_1_2_3_patch1-ProcStr-v1/TIER",
                         "/PD/CMSSW_1_2_3_HLT1-ProcStr-v1/TIER",
                         "/PD/CMSSW_1_2_3-ProcStr-v1/TIER",
                         "/PD/AcqEra-ProcStr-v1/TIER"}:
            self.assertFalse(policyObj._isPreRelease(dsetName))
        self.assertTrue(policyObj._isPreRelease("/PD/CMSSW_1_2_3_pre4-ProcStr-v1/TIER"))


if __name__ == '__main__':
    unittest.main()



import json

from WMCore.Services.Service import Service


class WMBS(Service):

    """
    API for dealing with retrieving information from PhEDEx DataService
    """

    def __init__(self, dict={}):
        """
        responseType will be either xml or json
        """

        dict.setdefault("accept_type", "application/json")
        dict.setdefault("content_type", "application/json")
        self.encoder = json.dumps
        self.decoder = json.loads

        Service.__init__(self, dict)

    def _getResult(self, callname, clearCache = True,
                   args = None, verb="GET", contentType = None):
        """
        _getResult_

        retrieve JSON/XML formatted information given the service name and the
        argument dictionaries

        TODO: Probably want to move this up into Service
        """
        result = ''
        # make base file name from call name.
        file = callname.replace("/", "_")
        if clearCache:
            self.clearCache(file, args, verb)

        # can't pass the decoder here since refreshCache wright to file
        f = self.refreshCache(file, callname, args, encoder = self.encoder,
                              verb = verb, contentType = contentType)
        result = f.read()
        f.close()
        result = self.decoder(result)

        return result

    def getJobSummaryFromCouchDB(self):
        """
        get the job status summary by request (workflow) from couchDB
        """
        args = {}
        callname = 'jobsummary'
        return self._getResult(callname, args = args, verb = "GET")

    def getSiteSummaryFromCouchDB(self):
        """
        returns number of complete jobs from the sites recorded in couchDB
        within the one hour period of calling. This might not the exact number
        since it could be gotten from the staled cache. (for performance reason)
        However, it should be close enough monitor sites' health.
        {'site': 'FNAL', 'complete': 1, 'success': 100, 'jobfailed':2}
        complete : job is completed but not accounted.
        """
        args = {}
        callname = 'jobstatebysite'
        return self._getResult(callname, args = args, verb = "GET")

    def getBatchJobStatus(self):
        """
        get the normalized job status from batch system
        """
        args = {}
        callname = 'batchjobstatus'
        return self._getResult(callname, args = args, verb = "GET")

    def getBatchJobStatusBySite(self):
        """
        get the normalized job status from batch system by sites
        """
        args = {}
        callname = 'batchjobstatusbysite'
        return self._getResult(callname, args = args, verb = "GET")

    def getResourceInfo(self, tableFormat = True):
        """
        """
        callname = 'listthresholdsforcreate'
        args = {'tableFormat': tableFormat}
        return self._getResult(callname, args = args)

    def getSiteList(self):
        """
        """
        callname = 'listsites'
        return self._getResult(callname)



#!/usr/bin/env python
"""
    _SandboxCreator_t_

    Tests sandbox production
"""

import os
import os.path
import pickle
import shutil
import subprocess
import tarfile
import tempfile
import unittest

import WMCore_t.WMSpec_t.TestWorkloads as TestWorkloads

import WMCore.WMRuntime.SandboxCreator as SandboxCreator
import WMCore.WMSpec.WMTask as WMTask


class SandboxCreator_t(unittest.TestCase):

    def testMakeSandbox(self):
        creator  = SandboxCreator.SandboxCreator()
        workload = TestWorkloads.twoTaskTree()
        tempdir  = tempfile.mkdtemp()
        # test that the existing path is deleted else it will crash as in issue #5130
        os.makedirs('%s/%s/WMSandbox' % (tempdir, workload.name()))
        boxpath  = creator.makeSandbox(tempdir, workload)

        # extract our sandbox to test it
        extractDir = tempfile.mkdtemp()
        tarHandle  = tarfile.open(boxpath, 'r:bz2')
        tarHandle.extractall( extractDir )

        self.fileExistsTest( extractDir + "/WMSandbox")
        self.fileExistsTest( extractDir + "/WMSandbox/WMWorkload.pkl")
        self.fileExistsTest( extractDir + "/WMSandbox/__init__.py")
        self.fileExistsTest( extractDir + "/WMSandbox/FirstTask/__init__.py")

        self.fileExistsTest( extractDir + "/WMSandbox/FirstTask/cmsRun1")
        self.fileExistsTest( extractDir + "/WMSandbox/FirstTask/stageOut1")
        self.fileExistsTest( extractDir + "/WMSandbox/FirstTask/cmsRun1/__init__.py")
        self.fileExistsTest( extractDir + "/WMSandbox/FirstTask/stageOut1/__init__.py")

        self.fileExistsTest( extractDir + "/WMSandbox/SecondTask/__init__.py")
        self.fileExistsTest( extractDir + "/WMSandbox/SecondTask/cmsRun2")
        self.fileExistsTest( extractDir + "/WMSandbox/SecondTask/stageOut2")
        self.fileExistsTest( extractDir + "/WMSandbox/SecondTask/cmsRun2/__init__.py")
        self.fileExistsTest( extractDir + "/WMSandbox/SecondTask/stageOut2/__init__.py")

        # make sure the sandbox is there
        self.fileExistsTest( extractDir + '/WMCore.zip')

        # Test that zipimport works on the dummy module that SandboxCreator inserts
        output = subprocess.check_output(['python', '-m', 'WMCore.ZipImportTestModule'],
                                         env={'PYTHONPATH': os.path.join(extractDir, 'WMCore.zip')})
        self.assertIn(b'ZIPIMPORTTESTOK', output)

        # make sure the pickled file is the same
        pickleHandle = open( extractDir + "/WMSandbox/WMWorkload.pkl", "rb")
        pickledWorkload = pickle.load( pickleHandle )
        self.assertEqual( workload.data, pickledWorkload )
        self.assertEqual( pickledWorkload.sandbox, boxpath )

        #TODO:This test will be deprecated when task.data.input.sandbox property is removed
        for task in workload.taskIterator():
            for t in task.nodeIterator():
                t = WMTask.WMTaskHelper(t)
                self.assertEqual(t.data.input.sandbox, boxpath)

        pickleHandle.close()

        pickledWorkload.section_("test_section")
        self.assertNotEqual( workload.data, pickledWorkload )
        shutil.rmtree( extractDir )
        shutil.rmtree( tempdir )

    def fileExistsTest(self, file, msg=None):
        if msg is None:
            msg = "Failed file existence test for (%s)" % file
        self.assertEqual(os.path.exists(file), True, msg)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""
_ReportIntegration_t_

Verify that the whole FWJR chain works correctly:
  CMSSW XML -> XMLParser -> Report -> Pickle -> UnPickle -> Accountant
"""

import unittest
import os
import xml.dom.minidom
import tempfile
import threading

import WMCore.WMBase
from WMQuality.TestInit import TestInit
from WMCore.DAOFactory import DAOFactory

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.WMBS.Workflow import Workflow

from WMComponent.JobAccountant.JobAccountantPoller import JobAccountantPoller
from WMCore.FwkJobReport.Report import Report
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile

class ReportIntegrationTest(unittest.TestCase):
    """
    _ReportIntegrationTest_

    """
    def setUp(self):
        """
        _setUp_

        Setup the database and WMBS for the test.
        """
        self.testInit = TestInit(__file__)
        self.testInit.setLogging()
        self.testInit.setDatabaseConnection()
        self.testInit.setSchema(customModules = ["WMComponent.DBS3Buffer",
                                                 "WMCore.WMBS"],
                                useDefault = False)

        myThread = threading.currentThread()
        self.daofactory = DAOFactory(package = "WMCore.WMBS",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        self.dbsfactory = DAOFactory(package = "WMComponent.DBS3Buffer",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        locationAction = self.daofactory(classname = "Locations.New")
        locationAction.execute(siteName = "site1", pnn = "T1_US_FNAL_Disk")

        inputFile = File(lfn = "/path/to/some/lfn", size = 10, events = 10,
                         locations = "T1_US_FNAL_Disk")
        inputFile.create()

        inputFileset = Fileset(name = "InputFileset")
        inputFileset.create()
        inputFileset.addFile(inputFile)
        inputFileset.commit()

        unmergedFileset = Fileset(name = "UnmergedFileset")
        unmergedFileset.create()

        mergedFileset = Fileset(name = "MergedFileset")
        mergedFileset.create()

        procWorkflow = Workflow(spec = "wf001.xml", owner = "Steve",
                                name = "TestWF", task = "/TestWF/None")
        procWorkflow.create()
        procWorkflow.addOutput("outputRECORECO", unmergedFileset)

        mergeWorkflow = Workflow(spec = "wf002.xml", owner = "Steve",
                                 name = "MergeWF", task = "/MergeWF/None")
        mergeWorkflow.create()
        mergeWorkflow.addOutput("Merged", mergedFileset)

        insertWorkflow = self.dbsfactory(classname = "InsertWorkflow")
        insertWorkflow.execute("TestWF", "/TestWF/None", 0, 0, 0, 0)
        insertWorkflow.execute("MergeWF", "/MergeWF/None", 0, 0, 0, 0)

        self.procSubscription = Subscription(fileset = inputFileset,
                                             workflow = procWorkflow,
                                             split_algo = "FileBased",
                                             type = "Processing")
        self.procSubscription.create()
        self.procSubscription.acquireFiles()

        self.mergeSubscription = Subscription(fileset = unmergedFileset,
                                             workflow = mergeWorkflow,
                                             split_algo = "WMBSMergeBySize",
                                             type = "Merge")
        self.mergeSubscription.create()

        self.procJobGroup = JobGroup(subscription = self.procSubscription)
        self.procJobGroup.create()
        self.mergeJobGroup = JobGroup(subscription = self.mergeSubscription)
        self.mergeJobGroup.create()

        self.testJob = Job(name = "testJob", files = [inputFile])
        self.testJob.create(group = self.procJobGroup)
        self.testJob["state"] = "complete"

        myThread = threading.currentThread()
        self.daofactory = DAOFactory(package = "WMCore.WMBS",
                                     logger = myThread.logger,
                                     dbinterface = myThread.dbi)
        self.stateChangeAction = self.daofactory(classname = "Jobs.ChangeState")
        self.setFWJRAction = self.daofactory(classname = "Jobs.SetFWJRPath")
        self.getJobTypeAction = self.daofactory(classname = "Jobs.GetType")
        locationAction = self.daofactory(classname = "Locations.New")
        locationAction.execute(siteName = "cmssrm.fnal.gov")

        self.stateChangeAction.execute(jobs = [self.testJob])

        self.tempDir = tempfile.mkdtemp()
        return

    def tearDown(self):
        """
        _tearDown_

        Clear out the database and the pickled report file.
        """
        self.testInit.clearDatabase()

        try:
            os.remove(os.path.join(self.tempDir, "ProcReport.pkl"))
            os.remove(os.path.join(self.tempDir, "MergeReport.pkl"))
        except Exception as ex:
            pass

        try:
            os.rmdir(self.tempDir)
        except Exception as ex:
            pass

        return

    def createConfig(self, workerThreads):
        """
        _createConfig_

        Create a config for the JobAccountant with the given number of worker
        threads.  This config needs to include information for connecting to the
        database as the component will create it's own database connections.
        These parameters are still pulled from the environment.
        """
        config = self.testInit.getConfiguration()
        self.testInit.generateWorkDir(config)

        config.section_("JobStateMachine")
        config.JobStateMachine.couchurl = os.getenv("COUCHURL")
        config.JobStateMachine.couchDBName = "report_integration_t"
        config.JobStateMachine.jobSummaryDBName = "report_integration_wmagent_summary_t"

        config.component_("JobAccountant")
        config.JobAccountant.pollInterval = 60
        config.JobAccountant.workerThreads = workerThreads
        config.JobAccountant.componentDir = os.getcwd()
        config.JobAccountant.logLevel = 'SQLDEBUG'

        config.component_("TaskArchiver")
        config.TaskArchiver.localWMStatsURL = "%s/%s" % (config.JobStateMachine.couchurl, config.JobStateMachine.jobSummaryDBName)
        return config

    def verifyJobSuccess(self, jobID):
        """
        _verifyJobSuccess_

        Verify that the metadata for a successful job is correct.  This will
        check the outcome, retry count and state.
        """
        testJob = Job(id = jobID)
        testJob.load()

        assert testJob["state"] == "success", \
               "Error: test job in wrong state: %s" % testJob["state"]
        assert testJob["retry_count"] == 0, \
               "Error: test job has wrong retry count: %s" % testJob["retry_count"]
        assert testJob["outcome"] == "success", \
               "Error: test job has wrong outcome: %s" % testJob["outcome"]

        return

    def verifyFileMetaData(self, jobID, fwkJobReportFiles):
        """
        _verifyFileMetaData_

        Verify that all the files that were output by a job made it into WMBS
        correctly.  Compare the contents of WMBS to the files in the frameworks
        job report.

        Note that fwkJobReportFiles is a list of DataStructs File objects.
        """
        testJob = Job(id = jobID)
        testJob.loadData()

        inputLFNs = []
        for inputFile in testJob["input_files"]:
            inputLFNs.append(inputFile["lfn"])

        for fwkJobReportFile in fwkJobReportFiles:
            outputFile = File(lfn = fwkJobReportFile["lfn"])
            outputFile.loadData(parentage = 1)

            assert outputFile["events"] == int(fwkJobReportFile["events"]), \
                   "Error: Output file has wrong events: %s, %s" % \
                   (outputFile["events"], fwkJobReportFile["events"])
            assert outputFile["size"] == int(fwkJobReportFile["size"]), \
                   "Error: Output file has wrong size: %s, %s" % \
                   (outputFile["size"], fwkJobReportFile["size"])

            for ckType in fwkJobReportFile["checksums"]:
                assert ckType in outputFile["checksums"], \
                       "Error: Output file is missing checksums: %s" % ckType
                assert outputFile["checksums"][ckType] == fwkJobReportFile["checksums"][ckType], \
                       "Error: Checksums don't match."

            assert len(fwkJobReportFile["checksums"]) == \
                   len(outputFile["checksums"]), \
                   "Error: Wrong number of checksums."

            jobType = self.getJobTypeAction.execute(jobID = jobID)
            if jobType == "Merge":
                assert str(outputFile["merged"]) == "True", \
                       "Error: Merge jobs should output merged files."
            else:
                assert outputFile["merged"] == fwkJobReportFile["merged"], \
                       "Error: Output file merged output is wrong: %s, %s" % \
                       (outputFile["merged"], fwkJobReportFile["merged"])

            assert len(outputFile["locations"]) == 1, \
                   "Error: outputfile should have one location: %s" % outputFile["locations"]
            assert list(outputFile["locations"])[0] == list(fwkJobReportFile["locations"])[0], \
                   "Error: wrong location for file."

            assert len(outputFile["parents"]) == len(inputLFNs), \
                   "Error: Output file has wrong number of parents."
            for outputParent in outputFile["parents"]:
                assert outputParent["lfn"] in inputLFNs, \
                       "Error: Unknown parent file: %s" % outputParent["lfn"]

            fwjrRuns = {}
            for run in fwkJobReportFile["runs"]:
                fwjrRuns[run.run] = run.lumis

            for run in outputFile["runs"]:
                assert run.run in fwjrRuns, \
                       "Error: Extra run in output: %s" % run.run

                for lumi in run:
                    assert lumi in fwjrRuns[run.run], \
                           "Error: Extra lumi: %s" % lumi

                    fwjrRuns[run.run].remove(lumi)

                if len(fwjrRuns[run.run]) == 0:
                    del fwjrRuns[run.run]

            assert len(fwjrRuns) == 0, \
                   "Error: Missing runs, lumis: %s" % fwjrRuns

            testJobGroup = JobGroup(id = testJob["jobgroup"])
            testJobGroup.loadData()
            jobGroupFileset = testJobGroup.output
            jobGroupFileset.loadData()

            assert outputFile["id"] in jobGroupFileset.getFiles(type = "id"), \
                   "Error: output file not in jobgroup fileset."

            if testJob["mask"]["FirstEvent"] == None:
                assert outputFile["first_event"] == 0, \
                       "Error: first event not set correctly: 0, %s" % \
                       outputFile["first_event"]
            else:
                assert testJob["mask"]["FirstEvent"] == outputFile["first_event"], \
                       "Error: last event not set correctly: %s, %s" % \
                       (testJob["mask"]["FirstEvent"], outputFile["first_event"])

        return

    def testReportHandling(self):
        """
        _testReportHandling_

        Verify that we're able to parse a CMSSW report, convert it to a Report()
        style report, pickle it and then have the accountant process it.
        """
        self.procPath = os.path.join(WMCore.WMBase.getTestBase(),
                                    "WMCore_t/FwkJobReport_t/CMSSWProcessingReport.xml")

        myReport = Report("cmsRun1")
        myReport.parse(self.procPath)

        # Fake some metadata that should be added by the stageout scripts.
        for fileRef in myReport.getAllFileRefsFromStep("cmsRun1"):
            fileRef.size = 1024
            fileRef.location = "cmssrm.fnal.gov"

        fwjrPath = os.path.join(self.tempDir, "ProcReport.pkl")
        cmsRunStep = myReport.retrieveStep("cmsRun1")
        cmsRunStep.status = 0
        myReport.setTaskName('/TestWF/None')
        myReport.persist(fwjrPath)

        self.setFWJRAction.execute(jobID = self.testJob["id"], fwjrPath = fwjrPath)

        pFile = DBSBufferFile(lfn = "/path/to/some/lfn", size = 600000, events = 60000)
        pFile.setAlgorithm(appName = "cmsRun", appVer = "UNKNOWN",
                           appFam = "RECO", psetHash = "GIBBERISH",
                           configContent = "MOREGIBBERISH")
        pFile.setDatasetPath("/bogus/dataset/path")
        #pFile.addRun(Run(1, *[45]))
        pFile.create()

        config = self.createConfig(workerThreads = 1)
        accountant = JobAccountantPoller(config)
        accountant.setup()
        accountant.algorithm()

        self.verifyJobSuccess(self.testJob["id"])
        self.verifyFileMetaData(self.testJob["id"], myReport.getAllFilesFromStep("cmsRun1"))

        inputFile = File(lfn = "/store/backfill/2/unmerged/WMAgentCommissioining10/MinimumBias/RECO/rereco_GR09_R_34X_V5_All_v1/0000/outputRECORECO.root")
        inputFile.load()
        self.testMergeJob = Job(name = "testMergeJob", files = [inputFile])
        self.testMergeJob.create(group = self.mergeJobGroup)
        self.testMergeJob["state"] = "complete"
        self.stateChangeAction.execute(jobs = [self.testMergeJob])

        self.mergePath = os.path.join(WMCore.WMBase.getTestBase(),
                                         "WMCore_t/FwkJobReport_t/CMSSWMergeReport.xml")

        myReport = Report("mergeReco")
        myReport.parse(self.mergePath)

        # Fake some metadata that should be added by the stageout scripts.
        for fileRef in myReport.getAllFileRefsFromStep("mergeReco"):
            fileRef.size = 1024
            fileRef.location = "cmssrm.fnal.gov"
            fileRef.dataset = {"applicationName": "cmsRun", "applicationVersion": "CMSSW_3_4_2_patch1",
                               "primaryDataset": "MinimumBias", "processedDataset": "Rereco-v1",
                               "dataTier": "RECO"}

        fwjrPath = os.path.join(self.tempDir, "MergeReport.pkl")
        myReport.setTaskName('/MergeWF/None')
        cmsRunStep = myReport.retrieveStep("mergeReco")
        cmsRunStep.status = 0
        myReport.persist(fwjrPath)

        self.setFWJRAction.execute(jobID = self.testMergeJob["id"], fwjrPath = fwjrPath)
        accountant.algorithm()

        self.verifyJobSuccess(self.testMergeJob["id"])
        self.verifyFileMetaData(self.testMergeJob["id"], myReport.getAllFilesFromStep("mergeReco"))

        return

if __name__ == "__main__":
    unittest.main()
